# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import os
import shutil
from collections.abc import Callable
from concurrent.futures import ThreadPoolExecutor, as_completed
from pathlib import Path

import tqdm
from compressed_tensors.entrypoints.convert.convert_file import (
    convert_file,
    validate_file,
)
from compressed_tensors.entrypoints.convert.converters import Converter
from compressed_tensors.entrypoints.convert.file_utils import (
    get_checkpoint_files,
    is_weights_file,
)
from compressed_tensors.entrypoints.convert.save_utils import (
    update_config,
    update_safetensors_index,
)
from loguru import logger


__all__ = ["convert_checkpoint", "exec_jobs"]


def convert_checkpoint(
    model_stub: str | os.PathLike,
    save_directory: str | os.PathLike,
    converter: Converter,
    max_workers: int = 1,
):
    """
    Convert a model checkpoint to compressed-tensors format without loading it up
    in memory, instead operating directly on the model safetensors files. This
    entrypoint operates on a model stub or folder containing weights saved in
    safetensors files, and updates the corresponding quantization_config field in
    the config.json. All additional files will be copied to new checkpoint.

    :param model_stub: huggingface model hub or path to local weights files
    :param save_directory: new checkpoint will be saved in this directory.
    :param max_workers: number of worker threads to process files with
    :param device: gpu device to accelerate quantization with
    :param converters: converter we wish to apply to the checkpoint,
        e.g. conversion of some layers from some format to compressed-tensors
    """
    # validate arguments
    model_files = get_checkpoint_files(model_stub)

    # 0. collect safetensors files, copy files
    validate_jobs = []
    convert_jobs = []
    for file_path, resolved_path in model_files.items():
        save_path = Path(save_directory) / file_path

        if file_path.endswith("safetensors"):
            validate_jobs.append((validate_file, resolved_path, converter))
            convert_jobs.append((convert_file, resolved_path, save_path, converter))

        else:
            if is_weights_file(file_path):
                logger.warning(f"Skip processing for weights file {file_path}")
            save_path.parent.mkdir(parents=True, exist_ok=True)
            logger.info(f"Copying {file_path} {save_path}")
            shutil.copyfile(resolved_path, save_path)

    # 1. validate quantizable tensors fail fast before long-running quantization
    exec_jobs(validate_jobs, max_workers, desc="Validating")

    # 2-5. quantize and compress weights
    total_size = 0
    weight_map = dict()
    convert_results = exec_jobs(convert_jobs, max_workers, desc="Converting")
    for _total_size, _weight_map in convert_results:
        total_size += _total_size
        weight_map.update(_weight_map)

    # 5. update config and safetensors index
    update_config(save_directory, converter)
    update_safetensors_index(save_directory, total_size, weight_map)


def exec_jobs(
    jobs: list[tuple[Callable, ...]], max_workers: int = 1, desc: str = "Executing Jobs"
) -> list:
    """
    Execute jobs in parallel, using ThreadPoolExecutor

    :param jobs: list of tuples, the first entry of which is the callable,
        and the remaining elements are the inputs args to the callable
    :param max_workers: number of workers to use
    :param desc: tqdm description
    """
    results = []
    with ThreadPoolExecutor(max_workers) as executor:
        futures = [executor.submit(*job) for job in jobs]
        for future in tqdm.tqdm(as_completed(futures), total=len(futures), desc=desc):
            results.append(future.result())

    return results



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

from copy import deepcopy

import torch
from compressed_tensors.quantization import (
    QuantizationConfig,
    apply_quantization_config,
    disable_quantization,
    enable_quantization,
)
from torch.nn import Linear


def test_quantization_enabled_disabled():
    inp = torch.randn(16)
    model = Linear(16, 16)
    quantized_model = deepcopy(model)
    apply_quantization_config(
        model=quantized_model,
        config=QuantizationConfig(
            config_groups=dict(W8A8=["Linear"]),
            quantization_status="calibration",
        ),
    )

    # run one calibration pass
    quantized_model(inp)

    model_output = model(inp)
    quantized_model_output = quantized_model(inp)

    # quantized and non quantized outputs should be different
    assert not torch.all(model_output == quantized_model_output)

    # disable quantization
    quantized_model.apply(disable_quantization)
    # check that quantized model now matches model output
    assert torch.all(model_output == quantized_model(inp))

    # re-enable quantization
    quantized_model.apply(enable_quantization)
    # check that quantized model matches original quantized output
    assert torch.all(quantized_model_output == quantized_model(inp))



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import math

import pytest
import torch
from compressed_tensors.quantization.lifecycle.forward import (
    _dequantize,
    _process_quantization,
    _quantize,
    _quantize_dequantize,
    fake_quantize,
    forward_quantize,
    set_forward_quantized,
)
from compressed_tensors.quantization.lifecycle.initialize import (
    initialize_module_for_quantization,
)
from compressed_tensors.quantization.quant_args import (
    QuantizationArgs,
    QuantizationStrategy,
)
from compressed_tensors.quantization.quant_config import QuantizationStatus
from compressed_tensors.quantization.utils.helpers import calculate_range
from torch.nn import Embedding, Linear


def make_dummy_g_idx(columns: int, group_size: int) -> torch.Tensor:
    perm = torch.randperm(columns)
    return torch.tensor([index // group_size for index in range(columns)])[perm]


def test_set_forward_quantized():
    layer = Linear(4, 4)
    func_forward = layer.forward.__func__

    # check that the forward call is overwritten
    set_forward_quantized(layer)
    assert not func_forward == layer.forward.__func__


def test_set_forward_quantized_embedding():
    """Test that set_forward_quantized works with Embedding modules"""
    embedding = Embedding(num_embeddings=10, embedding_dim=4)
    func_forward = embedding.forward.__func__

    # check that the forward call is overwritten
    set_forward_quantized(embedding)
    assert not func_forward == embedding.forward.__func__


def test_set_forward_quantized_embedding_no_quantization():
    """
    Test forward pass of Embedding when quantization is disabled or
    scheme is not set
    """
    embedding = Embedding(num_embeddings=10, embedding_dim=4)
    set_forward_quantized(embedding)

    input_indices = torch.tensor([0, 1, 2, 3])
    expected_output = torch.nn.functional.embedding(input_indices, embedding.weight)

    # Without quantization scheme, should behave like normal embedding
    output = embedding(input_indices)
    assert torch.allclose(output, expected_output)


def test_set_forward_quantized_embedding_with_weight_quantization(
    mock_per_tensor_calibration, create_quantization_scheme
):
    """Test forward pass with weight quantization on Embedding module"""
    num_bits = 8
    embedding = Embedding(num_embeddings=10, embedding_dim=4)
    embedding.weight.data *= 10

    quantization_scheme = create_quantization_scheme(
        targets=["*"],
        weights=QuantizationArgs(num_bits=num_bits, symmetric=True),
    )

    # initialize_module_for_quantization calls set_forward_quantized
    initialize_module_for_quantization(embedding, quantization_scheme)
    embedding.quantization_status = QuantizationStatus.CALIBRATION

    # Calibrate weights
    mock_per_tensor_calibration(embedding, "weight", value=embedding.weight.data)

    # Forward pass should quantize weights
    input_indices = torch.tensor([0, 1, 2, 3])
    output = embedding(input_indices)
    assert output.shape == (4, 4)

    # Output should be different from unquantized forward
    unquantized_output = torch.nn.functional.embedding(input_indices, embedding.weight)
    assert not torch.allclose(output, unquantized_output, atol=1e-3)


def test_set_forward_quantized_no_quantization():
    """Test forward pass when quantization is disabled or scheme is not set"""
    layer = Linear(4, 4)
    set_forward_quantized(layer)

    input_tensor = torch.randn(2, 4)
    expected_output = torch.nn.functional.linear(input_tensor, layer.weight, layer.bias)

    # Without quantization scheme, should behave like normal linear
    output = layer(input_tensor)
    assert torch.allclose(output, expected_output)


def test_set_forward_quantized_disabled():
    """Test forward pass when quantization_enabled is False"""
    layer = Linear(4, 4)
    set_forward_quantized(layer)

    # Set up quantization but disable it
    layer.quantization_enabled = False
    layer.quantization_scheme = torch.nn.Module()  # dummy scheme
    layer.quantization_status = QuantizationStatus.INITIALIZED

    input_tensor = torch.randn(2, 4)
    expected_output = torch.nn.functional.linear(input_tensor, layer.weight, layer.bias)

    # With quantization disabled, should behave like normal linear
    output = layer(input_tensor)
    assert torch.allclose(output, expected_output)


@pytest.mark.parametrize(
    "quantization_status",
    [
        QuantizationStatus.INITIALIZED,
        QuantizationStatus.CALIBRATION,
        QuantizationStatus.FROZEN,
    ],
)
def test_set_forward_quantized_with_input_activations(
    mock_per_tensor_calibration, create_quantization_scheme, quantization_status
):
    """Test forward pass with input activation quantization"""
    num_bits = 8
    layer = Linear(4, 4)
    layer.weight.data *= 10

    quantization_scheme = create_quantization_scheme(
        targets=["*"],
        input_activations=QuantizationArgs(num_bits=num_bits, symmetric=True),
    )

    # initialize_module_for_quantization calls set_forward_quantized
    initialize_module_for_quantization(layer, quantization_scheme)
    layer.quantization_status = quantization_status

    # Calibrate input activations
    input_tensor = torch.randn(2, 4)
    mock_per_tensor_calibration(layer, "input", value=input_tensor)

    # Forward pass should quantize inputs
    output = layer(input_tensor)
    assert output.shape == (2, 4)
    # Output should be different from unquantized forward
    unquantized_output = torch.nn.functional.linear(
        input_tensor, layer.weight, layer.bias
    )
    assert not torch.allclose(output, unquantized_output, atol=1e-3)


@pytest.mark.parametrize(
    "quantization_status",
    [
        QuantizationStatus.INITIALIZED,
        QuantizationStatus.CALIBRATION,
    ],
)
def test_set_forward_quantized_with_weight_quantization(
    mock_per_tensor_calibration, create_quantization_scheme, quantization_status
):
    """Test forward pass with weight quantization (non-FROZEN status)"""
    num_bits = 8
    layer = Linear(4, 4)
    layer.weight.data *= 10

    quantization_scheme = create_quantization_scheme(
        targets=["*"],
        weights=QuantizationArgs(num_bits=num_bits, symmetric=True),
    )

    # initialize_module_for_quantization calls set_forward_quantized
    initialize_module_for_quantization(layer, quantization_scheme)
    layer.quantization_status = quantization_status

    # Calibrate weights
    mock_per_tensor_calibration(layer, "weight", value=layer.weight.data)

    # Forward pass should quantize weights
    input_tensor = torch.randn(2, 4)
    output = layer(input_tensor)
    assert output.shape == (2, 4)


def test_set_forward_quantized_compressed_status(
    mock_per_tensor_calibration, create_quantization_scheme
):
    """Test that weight quantization is skipped when status is FROZEN"""
    num_bits = 8
    layer = Linear(4, 4)
    layer.weight.data *= 10

    quantization_scheme = create_quantization_scheme(
        targets=["*"],
        weights=QuantizationArgs(num_bits=num_bits, symmetric=True),
    )

    # initialize_module_for_quantization calls set_forward_quantized
    initialize_module_for_quantization(layer, quantization_scheme)
    layer.quantization_status = QuantizationStatus.COMPRESSED

    # Calibrate weights
    mock_per_tensor_calibration(layer, "weight", value=layer.weight.data)

    # Forward pass should NOT quantize weights due to FROZEN status
    input_tensor = torch.randn(2, 4)
    output = layer(input_tensor)
    expected_output = torch.nn.functional.linear(input_tensor, layer.weight, layer.bias)
    assert torch.allclose(output, expected_output)


def test_set_forward_quantized_with_output_activations(
    mock_per_tensor_calibration, create_quantization_scheme
):
    """Test forward pass with output activation quantization"""
    num_bits = 8
    layer = Linear(4, 4)
    layer.weight.data *= 10

    quantization_scheme = create_quantization_scheme(
        targets=["*"],
        output_activations=QuantizationArgs(num_bits=num_bits, symmetric=True),
    )

    # initialize_module_for_quantization calls set_forward_quantized
    initialize_module_for_quantization(layer, quantization_scheme)
    layer.quantization_status = QuantizationStatus.CALIBRATION

    # Need to calibrate output activations
    input_tensor = torch.randn(2, 4)
    output_sample = torch.nn.functional.linear(input_tensor, layer.weight, layer.bias)
    mock_per_tensor_calibration(layer, "output", value=output_sample)

    # Forward pass should quantize outputs
    output = layer(input_tensor)
    assert output.shape == (2, 4)


def test_set_forward_quantized_full_quantization(
    mock_per_tensor_calibration, create_quantization_scheme
):
    """Test forward pass with input, weight, and output quantization enabled"""
    num_bits = 8
    layer = Linear(4, 4)
    layer.weight.data *= 10

    quantization_scheme = create_quantization_scheme(
        targets=["*"],
        input_activations=QuantizationArgs(num_bits=num_bits, symmetric=True),
        weights=QuantizationArgs(num_bits=num_bits, symmetric=True),
        output_activations=QuantizationArgs(num_bits=num_bits, symmetric=True),
    )

    # initialize_module_for_quantization calls set_forward_quantized
    initialize_module_for_quantization(layer, quantization_scheme)
    layer.quantization_status = QuantizationStatus.CALIBRATION

    # Calibrate all components
    input_tensor = torch.randn(2, 4)
    mock_per_tensor_calibration(layer, "weight", value=layer.weight.data)
    mock_per_tensor_calibration(layer, "input", value=input_tensor)
    output_sample = torch.nn.functional.linear(input_tensor, layer.weight, layer.bias)
    mock_per_tensor_calibration(layer, "output", value=output_sample)

    # Forward pass should quantize all components
    output = layer(input_tensor)
    assert output.shape == (2, 4)
    # Should be significantly different from unquantized
    unquantized_output = torch.nn.functional.linear(
        input_tensor, layer.weight, layer.bias
    )
    assert not torch.allclose(output, unquantized_output, atol=1e-2)


@pytest.mark.parametrize("quantization_status", ["initialized", "calibration"])
def test_forward_quantize(
    mock_per_tensor_calibration, create_quantization_scheme, quantization_status
):
    num_bits = 8
    quantization_scheme = create_quantization_scheme(
        targets=["*"],
        weights=QuantizationArgs(num_bits=num_bits, symmetric=True),
        input_activations=QuantizationArgs(num_bits=num_bits, symmetric=True),
    )
    quantization_args = QuantizationArgs(num_bits=num_bits, symmetric=True)
    layer = Linear(4, 4)
    layer.weight.data *= 100

    dummy_tensor = torch.randn(8, 4)  # (num_tokens, num_features)
    layer.quantization_status = QuantizationStatus(quantization_status)

    # only calibration updates the scale and zero-point
    if layer.quantization_status == QuantizationStatus.INITIALIZED:
        # Init zp and scales
        initialize_module_for_quantization(layer, quantization_scheme)
        # mock weight calibration
        mock_per_tensor_calibration(layer, "weight", value=layer.weight.data)
        # call quant/dequant on weights
        out = forward_quantize(layer, layer.weight, "weight", quantization_args)
        assert torch.allclose(out, layer.weight.data, atol=0.2)
    elif layer.quantization_status == QuantizationStatus.CALIBRATION:
        # init zp/scales
        initialize_module_for_quantization(layer, quantization_scheme)
        # run weight and input calibration
        mock_per_tensor_calibration(layer, "weight", value=layer.weight.data)
        mock_per_tensor_calibration(layer, "input", value=dummy_tensor)
        # call quant/dequant on inputs
        out = forward_quantize(layer, dummy_tensor, "input", quantization_args)
        assert torch.allclose(out, dummy_tensor, atol=0.2)


@pytest.mark.parametrize(
    "num_bits,type,strategy,group_size,scale,zero_point,g_idx,global_scale",
    [
        (
            4,
            "int",
            QuantizationStrategy.TENSOR,
            None,
            torch.rand((1,)) * 0.01,
            torch.zeros((1,)),
            None,
            None,
        ),
        (
            4,
            "int",
            QuantizationStrategy.GROUP,
            128,
            torch.rand((512, 8)) * 0.01,
            torch.zeros((512, 8)),
            None,
            None,
        ),
        (
            4,
            "int",
            QuantizationStrategy.GROUP,
            128,
            torch.rand((512, 8)) * 0.01,
            torch.zeros((512, 8)),
            make_dummy_g_idx(1024, 128),
            None,
        ),
        (
            8,
            "float",
            QuantizationStrategy.TENSOR,
            None,
            torch.rand((1,)) * 0.01,
            torch.zeros((1,)),
            None,
            None,
        ),
        (
            8,
            "float",
            QuantizationStrategy.GROUP,
            128,
            torch.rand((512, 8)) * 0.01,
            torch.zeros((512, 8)),
            None,
            None,
        ),
        (
            8,
            "float",
            QuantizationStrategy.GROUP,
            128,
            torch.rand((512, 8)) * 0.01,
            torch.zeros((512, 8)),
            make_dummy_g_idx(1024, 128),
            None,
        ),
        (
            8,
            "int",
            QuantizationStrategy.GROUP,
            128,
            torch.rand((512, 8)) * 0.01,
            torch.zeros((512, 8)),
            None,
            None,
        ),
        (
            8,
            "int",
            QuantizationStrategy.GROUP,
            128,
            torch.rand((512, 8)) * 0.01,
            torch.zeros((512, 8)),
            make_dummy_g_idx(1024, 128),
            None,
        ),
    ],
)
def test_fake_quantize_2d(
    num_bits, type, strategy, group_size, scale, zero_point, g_idx, global_scale
):
    args = QuantizationArgs(
        num_bits=num_bits, type=type, strategy=strategy, group_size=group_size
    )

    x = torch.rand((512, 1024))
    fake_quantize(
        x=x,
        scale=scale,
        zero_point=zero_point,
        args=args,
        g_idx=g_idx,
        global_scale=global_scale,
    )  # note that reconstruction loss is bad for uncalibrated scales


def test_process_quantization_block_static():
    """
    Static block quantization (QuantizationStrategy.BLOCK) should split a 2D tensor
    into blocks, quantize each block, and reassemble without changing shape.
    """
    rows, cols = 8, 8
    bh, bw = 2, 4
    x = torch.randn(rows, cols)
    args = QuantizationArgs(
        num_bits=8,
        type="float",
        strategy=QuantizationStrategy.BLOCK,
        symmetric=True,
        dynamic=False,
        block_structure=[bh, bw],
    )
    num_rb = math.ceil(rows / bh)
    num_cb = math.ceil(cols / bw)
    scale = torch.rand(num_rb, num_cb) + 0.1
    zp = torch.zeros_like(scale)
    q_min, q_max = calculate_range(args, x.device)
    out = _process_quantization(
        x=x,
        scale=scale,
        zero_point=zp,
        args=args,
        do_quantize=True,
        do_dequantize=False,
        dtype=None,
        global_scale=None,
    )
    assert out.shape == x.shape
    # full fake-quantize roundtrip
    out2 = _process_quantization(
        x=x,
        scale=scale,
        zero_point=zp,
        args=args,
        do_quantize=True,
        do_dequantize=True,
        dtype=None,
        global_scale=None,
    )
    assert out2.shape == x.shape


@pytest.mark.parametrize(
    "rows,cols,block_height,block_width",
    [
        (4544, 768, 128, 128),  # Falcon-7B dimensions: 4544 = 64*71
        (100, 200, 128, 128),  # Both dimensions not divisible
        (256, 300, 128, 128),  # Only cols not divisible
        (300, 256, 128, 128),  # Only rows not divisible
        (127, 127, 128, 128),  # Both dimensions smaller than block size
        (1, 1, 128, 128),  # Minimal tensor
    ],
)
def test_process_quantization_block_non_divisible(
    rows, cols, block_height, block_width
):
    """
    Block quantization should handle tensor dimensions that are not divisible
    by the block size by padding internally.
    """
    x = torch.randn(rows, cols)
    args = QuantizationArgs(
        num_bits=8,
        type="float",
        strategy=QuantizationStrategy.BLOCK,
        symmetric=True,
        dynamic=False,
        block_structure=[block_height, block_width],
    )
    # Calculate number of blocks (with ceiling division for padding)
    num_rb = math.ceil(rows / block_height)
    num_cb = math.ceil(cols / block_width)
    scale = torch.rand(num_rb, num_cb) + 0.1
    zp = torch.zeros_like(scale)

    # Should NOT raise ValueError anymore
    out = _process_quantization(
        x=x,
        scale=scale,
        zero_point=zp,
        args=args,
        do_quantize=True,
        do_dequantize=False,
        dtype=None,
        global_scale=None,
    )
    # Output shape should match original input shape
    assert out.shape == x.shape, f"Expected {x.shape}, got {out.shape}"

    # Full fake-quantize roundtrip
    out2 = _process_quantization(
        x=x,
        scale=scale,
        zero_point=zp,
        args=args,
        do_quantize=True,
        do_dequantize=True,
        dtype=None,
        global_scale=None,
    )
    assert out2.shape == x.shape, f"Expected {x.shape}, got {out2.shape}"


@pytest.mark.parametrize(
    "rows,cols,block_height,block_width",
    [
        (100, 200, 128, 128),  # Both dimensions not divisible
        (256, 300, 128, 128),  # Only cols not divisible
        (300, 256, 128, 128),  # Only rows not divisible
        (127, 127, 128, 128),  # Both dimensions smaller than block size
    ],
)
def test_process_quantization_block_non_divisible_values(
    rows, cols, block_height, block_width
):
    """
    Verify that block quantization with non-divisible dimensions produces
    correct values. Using uniform input (ones) with scale=1.0 should result
    in zero quantization loss.
    """
    # Use uniform values - quantization with scale=1.0 should be lossless
    x = torch.ones(rows, cols)
    args = QuantizationArgs(
        num_bits=8,
        type="float",
        strategy=QuantizationStrategy.BLOCK,
        symmetric=True,
        dynamic=False,
        block_structure=[block_height, block_width],
    )
    num_rb = math.ceil(rows / block_height)
    num_cb = math.ceil(cols / block_width)
    # Use scale=1.0 for lossless quantization of values within FP8 range
    scale = torch.ones(num_rb, num_cb)
    zp = torch.zeros_like(scale)

    # Full fake-quantize roundtrip should preserve values exactly
    out = _process_quantization(
        x=x,
        scale=scale,
        zero_point=zp,
        args=args,
        do_quantize=True,
        do_dequantize=True,
        dtype=None,
        global_scale=None,
    )

    # Values should match input (no quantization loss for uniform values)
    assert out.shape == x.shape, f"Expected shape {x.shape}, got {out.shape}"
    assert torch.allclose(
        out, x, atol=1e-6
    ), f"Values mismatch: expected all ones, got min={out.min()}, max={out.max()}"

    # Test with a different uniform value
    x_val = torch.full((rows, cols), 0.5)
    out_val = _process_quantization(
        x=x_val,
        scale=scale,
        zero_point=zp,
        args=args,
        do_quantize=True,
        do_dequantize=True,
        dtype=None,
        global_scale=None,
    )
    assert torch.allclose(
        out_val, x_val, atol=1e-6
    ), f"Values mismatch for 0.5: got min={out_val.min()}, max={out_val.max()}"


@pytest.mark.parametrize(
    "num_bits,type,symmetric,global_scale",
    [
        (8, "int", True, None),
        (8, "int", False, None),
        (4, "int", True, None),
        (8, "float", True, None),
        (8, "float", True, torch.tensor([2.0])),
        (8, "int", False, torch.tensor([2.0])),
    ],
)
def test_quantize_dequantize_matches_sequential(
    num_bits, type, symmetric, global_scale
):
    """Verify that the fused _quantize_dequantize produces identical output
    to calling _quantize then _dequantize sequentially."""
    args = QuantizationArgs(
        num_bits=num_bits,
        type=type,
        symmetric=symmetric,
        strategy=QuantizationStrategy.TENSOR,
    )
    q_min, q_max = calculate_range(args, torch.device("cpu"))

    x = torch.randn(512, 1024)
    scale = torch.rand(1) * 0.01 + 0.001
    zero_point = None if symmetric else torch.tensor([3.0])

    # sequential: quantize then dequantize
    q = _quantize(
        x=x,
        scale=scale,
        zero_point=zero_point,
        q_min=q_min,
        q_max=q_max,
        args=args,
        global_scale=global_scale,
    )
    sequential_out = _dequantize(
        x_q=q,
        scale=scale,
        zero_point=zero_point,
        global_scale=global_scale,
    )

    # fused
    fused_out = _quantize_dequantize(
        x=x,
        scale=scale,
        zero_point=zero_point,
        q_min=q_min,
        q_max=q_max,
        args=args,
        global_scale=global_scale,
    )

    assert torch.equal(
        sequential_out, fused_out
    ), f"Mismatch: max diff = {(sequential_out - fused_out).abs().max().item()}"



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import contextlib
from collections.abc import Iterable
from typing import Literal

import torch
from compressed_tensors.offload.cache import OffloadCache
from compressed_tensors.offload.convert import from_accelerate, to_accelerate
from compressed_tensors.offload.dispatch import (  # noqa: F401
    dispatch_model,
    dispatch_with_map,
    get_device_map,
    offload_model,
    remove_dispatch,
)
from compressed_tensors.offload.dist_utils import (
    as_broadcastable,
    init_dist,
    is_distributed,
    is_rank0,
)
from compressed_tensors.offload.load import load_offloaded_model
from compressed_tensors.offload.module import offload_module, unwrap_offload_forward
from compressed_tensors.offload.utils import get_module_device, move_module_tensor
from compressed_tensors.utils.helpers import patch_attr


__all__ = [
    # dispatch models
    "offload_model",
    "dispatch_model",
    "remove_dispatch",
    "dispatch_with_map",
    "get_device_map",
    # accelerate conversion
    "load_offloaded_model",
    "from_accelerate",
    "to_accelerate",
    # control movement
    "disable_onloading",
    "disable_offloading",
    # manipulate parameters
    "update_offload_parameter",
    "get_execution_device",
    "get_offloaded_device",
    "register_offload_module",
    # manipulate forward
    "unwrap_offload_forward",
    # backwards compatibility: should be deprecated
    "align_modules",
    "align_module_device",
    # utilities
    "is_distributed",
    "is_rank0",
    "init_dist",
    "as_broadcastable",
]


@contextlib.contextmanager
def disable_offloading():
    """
    When offloading is disabled, onloaded tensors remain onloaded in memory until exit

    ```
    with OffloadCache.disable_offloading():
        ... = cache["weight"]
        ... = cache["weight"]  # cache hit
        ... = cache["weight"]  # cache hit

    # upon exit, all onloaded weights are released
    ```
    """
    with OffloadCache.disable_offloading():
        yield


@contextlib.contextmanager
def disable_onloading():
    """
    When onloading is disabled, tensors are not offloaded on access, and assignments do
    not trigger offloading. This is mostly used to disable device movement for debugging

    ```
    with OffloadCache.disable_onloading():
        tensor = ...
        cache["weight"] = tensor   # assignments do not trigger onloading
        cache["weight"] is tensor  # tensor remains offloaded
    ```
    """
    with OffloadCache.disable_onloading():
        yield


def update_offload_parameter(module: torch.nn.Module, name: str, data: torch.Tensor):
    """
    Update the offload and onload data of an existing parameter/buffer. Supports both
    parameters of both offloaded modules and non-offloaded modules.

    NOTE: This function does not guard against multiple processes writing to offload
    at the same time. It is the responsibility of the caller to ensure that, for any
    parameter/buffer, only one rank calls this function at a time.

    NOTE: This function does not update onloaded values across ranks. The caller is
    responsible for broadcasting any updates to other ranks, if they are onloaded.

    :param module: module containing the parameter to update
    :param name: name of module parameter to update
    :param data: tensor to update parameter with
    """
    if isinstance(module._parameters, OffloadCache):
        # | Component | Update Implementation       |
        # | --------- | --------------------------- |
        # | CPU       | Copy into shared cpu memory |
        # | Disk      | Write file to disk          |
        # | Device    | Copy into local device      |
        # | --------- | --------------------------- |
        # all implementations update onloaded data if applicable
        if name in module._parameters:
            cache = module._parameters
        elif name in module._buffers:
            cache = module._buffers
        else:
            raise AttributeError(f"{type(module)} has no attribute {name}")

        # triggers update if shapes match
        cache[name] = data

    else:
        with torch.no_grad():
            getattr(module, name).copy_(data)


def get_execution_device(
    module: torch.nn.Module, default: torch.device | None = None
) -> torch.device | Literal["disk"]:
    """
    Get the device which inputs should be moved to before module execution.

    :param module: module to check, may be offloaded
    :return: onload device of module
    """
    if isinstance(module._parameters, OffloadCache):
        return module._parameters.onload_device

    else:
        return get_module_device(module, default)


def get_offloaded_device(
    module: torch.nn.Module, default: torch.device | None = None
) -> torch.device | Literal["disk"]:
    """
    :param module: module to check
    :return: device module is offloaded to onto after forward pass
    """
    if isinstance(module._parameters, OffloadCache):
        return module._parameters.offload_device

    else:
        return get_module_device(module, default)


def register_offload_module(base: torch.nn.Module, name: str, module: torch.nn.Module):
    """
    Register a submodule with offloading if the parent module is offloaded

    :param base: module to attach submodule to
    :param name: name of submodule
    :param module: submodule to attach
    """
    cache = base._parameters
    if isinstance(cache, OffloadCache):
        offload_module(module, cache.onload_device, cache.offload_device)

    base.register_module(name, module)


""" Implemented for backwards compatibility """


@contextlib.contextmanager
def align_modules(
    modules: torch.nn.Module | Iterable[torch.nn.Module],
    execution_device: torch.device | None = None,
):
    """
    Context manager for onloading modules to a device, and disabling onload and offload
    attempts triggered by forward calls. Used for sequential onloading of layers

    :param modules: `torch.nn.Module` or iterable of `torch.nn.Module`s to onload
    :param execution_device: device to onload to
    """
    with contextlib.ExitStack() as stack:
        for module in modules:
            stack.enter_context(align_module_device(module, execution_device))
        yield


@contextlib.contextmanager
def align_module_device(
    module: torch.nn.Module, execution_device: torch.device | None = None
):
    """
    Context manager that moves a module's parameters to the specified execution device.

    :param module: Module with parameters to align
    :param execution_device: If provided, overrides the module's execution device
        within the context. Otherwise, use hook execution device or pass
    """

    if isinstance(module._parameters, OffloadCache):
        assert isinstance(module._buffers, OffloadCache)
        with module._parameters.disable_offloading():
            if execution_device is not None:
                with (
                    patch_attr(module._parameters, "onload_device", execution_device),
                    patch_attr(module._buffers, "onload_device", execution_device),
                ):
                    yield
            else:
                yield

    else:
        original_device = {}
        for name, param in module.named_parameters(recurse=False):
            original_device[name] = param.device
            move_module_tensor(module, name, execution_device)

        try:
            yield
        finally:
            for name, param in module.named_parameters(recurse=False):
                device = original_device[name]
                move_module_tensor(module, name, device)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import pytest
import torch
import torch.nn as nn
from compressed_tensors.compressors.base import compress_module, decompress_module
from compressed_tensors.config import CompressionFormat
from compressed_tensors.quantization import (
    ActivationOrdering,
    initialize_module_for_quantization,
    preset_name_to_scheme,
)
from compressed_tensors.utils import get_direct_state_dict


def _run_compress_decompress(scheme_name, expected_format, actorder, device):
    if device == "cuda" and not torch.cuda.is_available():
        pytest.skip("CUDA not available")

    # 1. Initialize module for quantization using a preset scheme.
    # Use 256x256 to avoid degenerate scale shapes (e.g. (N,1) for FP8_BLOCK)
    # that trip up block-vs-channel strategy inference in dequantize.
    # Start in bfloat16 so NVFP4 decompression (which returns bfloat16) preserves dtype.
    module = nn.Linear(256, 256, bias=False).to(dtype=torch.bfloat16, device=device)
    scheme = preset_name_to_scheme(scheme_name, ["Linear"])
    if actorder is not None:
        scheme.weights.actorder = actorder
    initialize_module_for_quantization(module, scheme)

    with torch.no_grad():
        for name, param in list(module.named_parameters()):
            param.fill_(1)

    # Record pre-compression state dict shapes and dtypes.
    # Filter out None entries (e.g. bias=None when bias=False).
    pre_state = {
        name: (tensor.shape, tensor.dtype)
        for name, tensor in get_direct_state_dict(module).items()
        if tensor is not None
    }

    # 2. Compress the module and verify the inferred quantization format.
    compress_module(module)
    assert module.quantization_scheme.format == expected_format

    # 3. Decompress the module and verify shapes and dtypes are restored.
    decompress_module(module)

    post_state_dict = get_direct_state_dict(module)
    for name, tensor in post_state_dict.items():
        if name in pre_state:
            pre_shape, pre_dtype = pre_state[name]
            assert tensor.shape == pre_shape
            assert tensor.dtype == pre_dtype


@pytest.mark.parametrize(
    "scheme_name,expected_format,actorder",
    [
        ("UNQUANTIZED", CompressionFormat.dense, None),
        ("W8A16", CompressionFormat.pack_quantized, None),
        ("W4A16", CompressionFormat.pack_quantized, None),
        ("W4A16", CompressionFormat.pack_quantized, ActivationOrdering.GROUP),
        ("W4A16_ASYM", CompressionFormat.pack_quantized, None),
        ("W4A16_ASYM", CompressionFormat.pack_quantized, ActivationOrdering.GROUP),
        ("W8A8", CompressionFormat.int_quantized, None),
        ("W4A8", CompressionFormat.int_quantized, None),
        ("W4AFP8", CompressionFormat.int_quantized, None),
        ("FP8", CompressionFormat.float_quantized, None),
        ("FP8_DYNAMIC", CompressionFormat.float_quantized, None),
        ("FP8_BLOCK", CompressionFormat.float_quantized, None),
        ("NVFP4A16", CompressionFormat.nvfp4_pack_quantized, None),
        ("NVFP4", CompressionFormat.nvfp4_pack_quantized, None),
    ],
)
@pytest.mark.parametrize("device", ["cpu", "meta", "cuda"])
def test_compress_decompress_module(scheme_name, expected_format, actorder, device):
    _run_compress_decompress(scheme_name, expected_format, actorder, device)


@pytest.mark.xfail(
    reason="MXFP4 decompression is not implemented",
    raises=NotImplementedError,
)
@pytest.mark.parametrize(
    "scheme_name,expected_format",
    [
        ("MXFP4A16", CompressionFormat.mxfp4_pack_quantized),
        ("MXFP4", CompressionFormat.mxfp4_pack_quantized),
    ],
)
@pytest.mark.parametrize("device", ["cpu", "meta", "cuda"])
def test_compress_decompress_module_mxfp4(scheme_name, expected_format, device):
    _run_compress_decompress(scheme_name, expected_format, None, device)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

from collections import Counter

import pytest
import torch
from compressed_tensors.offload import (
    disable_offloading,
    disable_onloading,
    offload_model,
)
from compressed_tensors.transform import (
    TransformArgs,
    TransformBase,
    TransformConfig,
    TransformScheme,
    apply_transform_config,
)
from tests.test_transform.conftest import TransformableModel
from tests.testing_utils import requires_gpu


@requires_gpu
@pytest.mark.parametrize("type", ("hadamard", "random-hadamard"))
@pytest.mark.parametrize("randomize", (True, False))
@pytest.mark.parametrize("requires_grad", (True, False))
# @pytest.mark.parametrize("offload", (True, False))
@pytest.mark.parametrize("offload", (True,))
def test_memory_sharing(type, randomize, requires_grad, offload):
    # load model (maybe with offloading)
    model = TransformableModel(2, 2, 4, 4, 8, 8)
    if offload:
        offload_model(model, torch.device("cuda"))

    # add transforms to model
    config = TransformConfig(
        config_groups={
            "": TransformScheme(
                type=type,
                randomize=randomize,
                requires_grad=requires_grad,
                apply=[
                    TransformArgs(targets="Linear", location="input"),
                    TransformArgs(targets="Linear", location="output"),
                ],
            )
        }
    )
    apply_transform_config(model, config)

    for context in disable_onloading, disable_offloading:
        with context():
            weights = [
                m.weight for m in model.modules() if isinstance(m, TransformBase)
            ]
            weight_to_count = Counter(weights)
            size_to_weight = {weight.size(0): weight for weight in weight_to_count}

            assert len(weight_to_count) == len(size_to_weight) == 3
            assert weight_to_count[size_to_weight[2]] == 3
            assert weight_to_count[size_to_weight[4]] == 4
            assert weight_to_count[size_to_weight[8]] == 3



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import torch
from compressed_tensors.compressors.mxfp8 import MXFP8QuantizationCompressor
from compressed_tensors.quantization import (
    QuantizationArgs,
    QuantizationScheme,
    QuantizationStrategy,
)
from compressed_tensors.quantization.utils.helpers import calculate_qparams


def test_mxfp8_compress_decompress():
    """
    Test MXFP8 compress/decompress round-trip with group strategy
    and group_size=32. Verifies weights survive the cycle (lossy but close).
    """
    rows, cols = 512, 1024
    group_size = 32
    num_groups = cols // group_size

    quant_args = QuantizationArgs(
        num_bits=8,
        type="float",
        strategy=QuantizationStrategy.GROUP,
        group_size=group_size,
        scale_dtype=torch.uint8,
        zp_dtype=torch.uint8,
        symmetric=True,
    )

    weight = torch.randn((rows, cols))

    # Compute scales using calculate_qparams (which generates MX scales)
    reshaped = weight.reshape(rows, num_groups, group_size)
    min_vals = reshaped.amin(dim=-1)
    max_vals = reshaped.amax(dim=-1)
    scale, zp = calculate_qparams(min_vals, max_vals, quant_args)

    scheme = QuantizationScheme(
        targets=["Linear"],
        weights=quant_args,
    )

    # Build per-module state dict (local names, no module prefix)
    module_sd = {
        "weight": weight,
        "weight_scale": scale,
        "weight_zero_point": zp,
    }

    # Compress
    compressed = MXFP8QuantizationCompressor.compress(module_sd, scheme=scheme)

    # Check compressed weight is FP8
    assert compressed["weight"].dtype == torch.float8_e4m3fn

    # Check scale is stored as uint8 (E8M0 exponent format)
    assert compressed["weight_scale"].dtype == torch.uint8

    # Decompress
    decompressed = MXFP8QuantizationCompressor.decompress(compressed, scheme=scheme)

    # Check shapes match
    assert decompressed["weight"].shape == weight.shape

    # FP8 quantization is lossy, but should be reasonably close
    assert torch.allclose(decompressed["weight"].float(), weight, atol=0.1, rtol=0.1)


def test_mxfp8_scale_roundtrip():
    """
    Test that E8M0 scale encoding/decoding is lossless for power-of-2 scales.
    """
    rows, cols = 128, 256
    group_size = 32
    num_groups = cols // group_size

    quant_args = QuantizationArgs(
        num_bits=8,
        type="float",
        strategy=QuantizationStrategy.GROUP,
        group_size=group_size,
        scale_dtype=torch.uint8,
        zp_dtype=torch.uint8,
        symmetric=True,
    )

    weight = torch.randn((rows, cols))

    reshaped = weight.reshape(rows, num_groups, group_size)
    min_vals = reshaped.amin(dim=-1)
    max_vals = reshaped.amax(dim=-1)
    scale, zp = calculate_qparams(min_vals, max_vals, quant_args)

    scheme = QuantizationScheme(
        targets=["Linear"],
        weights=quant_args,
    )

    module_sd = {
        "weight": weight,
        "weight_scale": scale,
        "weight_zero_point": zp,
    }

    compressed = MXFP8QuantizationCompressor.compress(module_sd, scheme=scheme)

    # E8M0 encoded scale
    e8m0_scale = compressed["weight_scale"]
    assert e8m0_scale.dtype == torch.uint8

    # Decode: 2^(exp - 127)
    scale_exp = e8m0_scale.to(torch.int32) - 127
    decoded_scale = 2.0 ** scale_exp.to(torch.float32)

    # The original scale after floor(log2) should round-trip exactly
    expected_scale = 2.0 ** torch.floor(torch.log2(scale)).to(torch.float32)
    assert torch.allclose(decoded_scale, expected_scale)


def test_mxfp8_can_compress():
    """Test that can_compress matches MXFP8 signature correctly."""
    import torch.nn as nn

    mxfp8_scheme = QuantizationScheme(
        targets=["Linear"],
        weights=QuantizationArgs(
            num_bits=8,
            type="float",
            strategy=QuantizationStrategy.GROUP,
            group_size=32,
            scale_dtype=torch.uint8,
        ),
    )
    assert MXFP8QuantizationCompressor.can_compress(nn.Linear, mxfp8_scheme) is True

    # Non-MXFP8: group_size != 32
    non_mxfp8_scheme = QuantizationScheme(
        targets=["Linear"],
        weights=QuantizationArgs(
            num_bits=8,
            type="float",
            strategy=QuantizationStrategy.GROUP,
            group_size=128,
        ),
    )
    assert (
        MXFP8QuantizationCompressor.can_compress(nn.Linear, non_mxfp8_scheme) is False
    )

    # Non-MXFP8: int type
    int_scheme = QuantizationScheme(
        targets=["Linear"],
        weights=QuantizationArgs(
            num_bits=8,
            type="int",
            strategy=QuantizationStrategy.GROUP,
            group_size=32,
            scale_dtype=torch.uint8,
        ),
    )
    assert MXFP8QuantizationCompressor.can_compress(nn.Linear, int_scheme) is False



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import torch
from compressed_tensors.offload.cache.base import OffloadCache
from compressed_tensors.offload.utils import send_tensors


class CPUCache(OffloadCache):
    """
    Handles offloading and onloading tensors from/to cpu memory
    """

    offload_device = torch.device("cpu")

    def __init__(self, onload_device: torch.device | str, offload_device=None):
        super().__init__(onload_device, offload_device=offload_device)

    def onload(self, offloaded: torch.Tensor | None) -> torch.Tensor | None:
        """
        Onload a tensor from cpu to device

        :param key: cpu tensor to onload
        :return: device tensor
        """
        return send_tensors(offloaded, device=self.onload_device, copy=False)

    def offload(self, tensor: torch.Tensor | None) -> torch.Tensor | None:
        """
        Offload a tensor from any device to cpu

        :param value: tensor on any device
        :return: tensor on cpu
        """
        return send_tensors(tensor, device=self.offload_device, copy=False)

    @torch.no_grad()
    def update_offload(self, offloaded: torch.Tensor, data: torch.Tensor | None):
        """
        Update the offloaded cpu value with new data

        :param offloaded: cpu tensor to update
        :param data: new data to copy from
        """
        if data is not None:
            offloaded.copy_(data)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

from collections import OrderedDict
from copy import deepcopy

import torch
from compressed_tensors.modeling import (
    initialize_hooked_attention,
    initialize_hooked_kv_cache,
)
from compressed_tensors.offload import update_offload_parameter
from compressed_tensors.quantization.lifecycle.initialize import (
    initialize_module_for_quantization,
    is_attention_module,
)
from compressed_tensors.quantization.quant_args import QuantizationArgs
from compressed_tensors.quantization.quant_config import (
    QuantizationConfig,
    QuantizationStatus,
)
from compressed_tensors.quantization.quant_scheme import QuantizationScheme
from compressed_tensors.utils.match import (
    is_narrow_match,
    match_named_modules,
    match_targets,
)
from compressed_tensors.utils.safetensors_load import get_safetensors_folder
from loguru import logger
from safetensors import safe_open
from torch.nn import Module


__all__ = [
    "load_pretrained_quantization_parameters",
    "apply_quantization_config",
]

from compressed_tensors.quantization.utils.helpers import is_module_quantized
from compressed_tensors.utils.safetensors_load import (
    get_quantization_parameter_to_path_mapping,
)


def load_pretrained_quantization_parameters(
    model: Module,
    model_name_or_path: str | None = None,
    load_weight_qparams: bool = False,
):
    """
    Loads the quantization parameters (scale and zero point) from model_name_or_path to
    a model that has already been initialized with a quantization config.

    NOTE: Will always load inputs/output parameters. Will conditioanlly load weight
    parameters, if load_weight_qparams is set to True.

    :param model: model to load pretrained quantization parameters to
    :param model_name_or_path: Hugging Face stub or local folder containing a quantized
        model, which is used to load quantization parameters
    :param load_weight_qparams: whether or not the weight quantization parameters
        should be loaded
    """
    model_path = get_safetensors_folder(model_name_or_path)
    mapping = get_quantization_parameter_to_path_mapping(model_path)

    for name, submodule in model.named_modules():
        if not is_module_quantized(submodule):
            continue
        if submodule.quantization_scheme.input_activations is not None:
            base_name = "input"
            _load_quant_args_from_mapping(
                base_name=base_name,
                module_name=name,
                module=submodule,
                mapping=mapping,
            )
        if submodule.quantization_scheme.output_activations is not None:
            base_name = "output"
            _load_quant_args_from_mapping(
                base_name=base_name,
                module_name=name,
                module=submodule,
                mapping=mapping,
            )

        if load_weight_qparams and submodule.quantization_scheme.weights:
            base_name = "weight"
            _load_quant_args_from_mapping(
                base_name=base_name,
                module_name=name,
                module=submodule,
                mapping=mapping,
            )


def apply_quantization_config(
    model: Module, config: QuantizationConfig | None, run_compressed: bool = False
):
    """
    Initializes the model for quantization in-place based on the given config.
    Optionally coverts quantizable modules to compressed_linear modules

    :param model: model to apply quantization config to
    :param config: quantization config
    :param run_compressed: Whether the model will be run in compressed mode or
        decompressed fully on load
    """
    config = deepcopy(config)
    if config is None:  # see PR #180
        return dict()

    # force zero points during initialization
    # TODO: remove zero points from initialization
    force_zero_point = config.quantization_status < QuantizationStatus.COMPRESSED

    # apply and initialize kv cache quantization
    if config.kv_cache_scheme is not None:
        _apply_kv_cache_scheme(
            model, config.kv_cache_scheme, config.quantization_status
        )

    # build mapping of targets to schemes for easier matching
    # use ordered dict to preserve target ordering in config
    target_to_scheme = OrderedDict()
    for scheme in config.config_groups.values():
        for target in scheme.targets:
            target_to_scheme[target] = scheme

    # mark appropriate layers for quantization by setting their quantization schemes
    for name, submodule in match_named_modules(
        model, target_to_scheme, config.ignore, warn_on_fail=True
    ):
        # mark modules to be quantized by adding
        # quant scheme to the matching layers
        matched_targets = match_targets(name, submodule, target_to_scheme)
        scheme = _scheme_from_targets(target_to_scheme, matched_targets, name)
        # target matched - add layer and scheme to target list
        submodule.quantization_scheme = scheme

        if is_attention_module(submodule) and is_narrow_match(
            model, scheme.targets, name
        ):
            initialize_hooked_attention(model, submodule)

        initialize_module_for_quantization(submodule, force_zero_point=force_zero_point)
        submodule.quantization_status = config.quantization_status


def _apply_kv_cache_scheme(
    model: torch.nn.Module,
    kv_cache_scheme: QuantizationArgs,
    status: QuantizationStatus,
):
    if not kv_cache_scheme.symmetric:
        raise logger.warning("vLLM does not support asymmetric kv cache quantization")

    # applies and initializes kv cache quantization
    # this step cannot come after attention apply/initialize
    # otherwise it will override the attention qparams
    scheme = QuantizationScheme(
        targets=[".*self_attn$"],  # is never read in practice
        input_activations=kv_cache_scheme,
    )
    for submodule in model.modules():
        if is_attention_module(submodule):
            submodule.quantization_scheme = scheme
            initialize_hooked_kv_cache(model, submodule)
            initialize_module_for_quantization(submodule, force_zero_point=False)
            submodule.quantization_status = status


def _load_quant_args_from_mapping(
    base_name: str, module_name: str, module: Module, mapping: dict
):
    # TODO: skip update and just register here, don't do it in initialize
    """
    Loads scale and zero point from a state_dict into the specified module

    :param base_name: quantization target, one of: weights, input_activations or
    output_activations
    :param module_name: pytorch module name to look up in state_dict
    :module: pytorch module associated with module_name
    :mapping: mapping to search fetch paths on disk for a given parameter
    """
    scale_name = f"{base_name}_scale"
    zp_name = f"{base_name}_zero_point"
    g_idx_name = f"{base_name}_g_idx"

    state_dict_scale_path = mapping.get(f"{module_name}.{scale_name}", None)
    state_dict_zp_path = mapping.get(f"{module_name}.{zp_name}", None)
    state_dict_g_idx_path = mapping.get(f"{module_name}.{g_idx_name}", None)

    if state_dict_g_idx_path is not None:
        with safe_open(state_dict_g_idx_path, framework="pt", device="cpu") as f:
            state_dict_g_idx = f.get_tensor(f"{module_name}.{g_idx_name}")

        update_offload_parameter(module, g_idx_name, state_dict_g_idx)

    if state_dict_scale_path is not None:
        # module is quantized
        with safe_open(state_dict_scale_path, framework="pt", device="cpu") as f:
            state_dict_scale = f.get_tensor(f"{module_name}.{scale_name}")

        update_offload_parameter(module, scale_name, state_dict_scale)

        if state_dict_zp_path is None:
            # fill in zero point for symmetric quantization
            state_dict_zp = torch.zeros_like(state_dict_scale, device="cpu")
        else:
            with safe_open(state_dict_zp_path, framework="pt", device="cpu") as f:
                state_dict_zp = f.get_tensor(f"{module_name}.{zp_name}")

        update_offload_parameter(module, zp_name, state_dict_zp)


def _scheme_from_targets(
    target_to_scheme: OrderedDict[str, QuantizationScheme],
    targets: list[str],
    name: str,
) -> QuantizationScheme:
    # return the first scheme (the prioritized one,
    # since the order of target_to_scheme matters)
    return target_to_scheme[targets[0]]



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import json
import os
from typing import Optional

import compressed_tensors
import torch
from compressed_tensors.base import (
    COMPRESSION_VERSION_NAME,
    QUANTIZATION_CONFIG_NAME,
    QUANTIZATION_METHOD,
    QUANTIZATION_METHOD_NAME,
    SPARSITY_CONFIG_NAME,
    TRANSFORM_CONFIG_NAME,
)
from compressed_tensors.compressors.base import compress_module, decompress_module
from compressed_tensors.compressors.format import infer_model_format
from compressed_tensors.config import CompressionFormat
from compressed_tensors.quantization import QuantizationConfig, QuantizationStatus
from compressed_tensors.quantization.utils.helpers import is_module_quantized
from compressed_tensors.transform import TransformConfig
from loguru import logger
from tqdm import tqdm
from transformers import CompressedTensorsConfig
from transformers.file_utils import CONFIG_NAME


__all__ = ["ModelCompressor"]


class ModelCompressor:
    """
    Orchestrates compression and decompression of a quantized model.

    Compression Lifecycle
        - model.save_pretrained_wrapper(quantized_path)
            - compressor = ModelCompressor.from_pretrained_model(model)
            - compressor.compress_model(model)
            - model.save_pretrained(quantized_path)
            - compressor.update_config(quantized_path)

    Decompression Lifecycle
        - model = AutoModelForCausalLM.from_pretrained(quantized_path)
            - CompressedTensorsHfQuantizer.__init__
                - compressor = ModelCompressor.from_compression_config(ct_config)
            - CompressedTensorsHfQuantizer._process_model_before_weight_loading
                - apply_quantization_config(model, ct_config.quantization_config)
                - compressor.compress_model(model)
            - CompressedTensorsHfQuantizer._process_model_after_weight_loading
                - if run_compressed == False: compressor.decompress_model(model)
    """

    # these attributes are used by `CompressedTensorsHfQuantizer` to apply configs
    # during `_process_model_before_weight_loading`
    quantization_config: QuantizationConfig | None
    transform_config: TransformConfig | None
    force_compression_format: CompressionFormat | None

    @classmethod
    def from_compression_config(cls, compression_config: CompressedTensorsConfig):
        """
        HFQuantizer uses this entrypoint to load quantized models by passing an
        instance of `CompressedTensorsConfig`

        :param compression_config: instance of `CompressedTensorsConfig` containing
            quantized model information
        """
        # vLLM Cutlass24 implementation is no longer supported
        if not isinstance(compression_config, CompressedTensorsConfig):
            raise ValueError(
                f"Support for compression config of type {type(compression_config)} "
                "is no longer supported. If you are attempting to use a Sparse24 "
                "model, note that the Sparse24 format is not longer supported by "
                "as of `compressed-tensors>0.14.0`"
            )

        # transform_config is added by transformers#42887
        q_config = compression_config.quantization_config
        t_config = getattr(compression_config, "transform_config", None)

        return cls(quantization_config=q_config, transform_config=t_config)

    @classmethod
    def from_pretrained_model(
        cls,
        model: torch.nn.Module,
        sparsity_config_or_format: Optional[object] = None,
        quantization_format: Optional[str] = None,
    ):
        """
        LLM Compressor uses this entrypoint to compress models before saving
        Given a path to a model config, extract the sparsity and/or quantization
        configs and load a ModelCompressor.

        Note: passing `sparsity_config_or_format` is no longer supported

        :param model: model with quantization config applied
        :param quantization_format: string corresponding to a quantization format
            that should be applied to the entire model, overrides inferred formats
            for all quantized modules
        """
        if sparsity_config_or_format is not None:
            logger.warning("Passing sparsity config or format is no longer supported")

        # reconstruct qconfig from qschemes that are attached to the model
        quantization_config = QuantizationConfig.from_pretrained(model)
        transform_config = getattr(model, TRANSFORM_CONFIG_NAME, None)

        # update quantization config format for better UX when reading config.json
        if quantization_config is not None:
            quantization_config.format = infer_model_format(model, quantization_format)

        return cls(
            quantization_config=quantization_config,
            transform_config=transform_config,
            force_compression_format=quantization_format,
        )

    def __init__(
        self,
        quantization_config: Optional[QuantizationConfig] = None,
        transform_config: Optional[TransformConfig] = None,
        force_compression_format: Optional[str] = None,
    ):
        self.quantization_config = quantization_config
        self.transform_config = transform_config
        self.force_compression_format = (
            CompressionFormat(force_compression_format)
            if force_compression_format is not None
            else None
        )

    def compress_model(self, model: torch.nn.Module) -> None:
        """
        Compress the model's parameters in memory

        :param model: model whose parameters should be compressed in place
        """
        # compress modules
        modules = model.named_modules(remove_duplicate=True)
        for _, module in tqdm(list(modules), desc="Compressing model"):
            if is_module_quantized(module):
                compress_module(module, self.force_compression_format)

        # update config status to reflect compression
        if self.quantization_config is not None:
            self.quantization_config.quantization_status = QuantizationStatus.COMPRESSED

        # attempting to perform forward passes with a compressed model
        # will cause to the model to decompress. This allows for generation
        # without requiring CT to handle quantized kernels
        self.add_decompress_hook(model)

    def decompress_model(self, model: torch.nn.Module) -> None:
        """
        Decompress the model's parameters in memory

        :param model: model whose parameters should be decompressed in place
        """
        # decompress modules
        modules = model.named_modules(remove_duplicate=True)
        for _, module in tqdm(list(modules), desc="Decompressing model"):
            if is_module_quantized(module):
                decompress_module(module, self.force_compression_format)

        # update config status to reflect decompression
        if self.quantization_config is not None:
            self.quantization_config.quantization_status = (
                QuantizationStatus.DECOMPRESSED
            )

        # decompression hook is no longer necessary
        self.remove_decompression_hook(model)

    def update_config(self, save_directory: str) -> None:
        """
        Update the model config located at save_directory with compression configs

        :param save_directory: path to a folder containing a HF model config
        """
        if not any((self.quantization_config, self.transform_config)):
            return

        config_file_path = os.path.join(save_directory, CONFIG_NAME)
        if os.path.exists(config_file_path):
            with open(config_file_path, "r") as file:
                config_data = json.load(file)
        else:
            config_data = {}

        qconfig_data = (
            self.quantization_config.model_dump(exclude=["quant_method"])
            if self.quantization_config is not None
            else {}
        )
        tconfig_data = (
            self.transform_config.model_dump()
            if self.transform_config is not None
            else {}
        )

        config_data[QUANTIZATION_CONFIG_NAME] = {
            COMPRESSION_VERSION_NAME: compressed_tensors.__version__,
            QUANTIZATION_METHOD_NAME: QUANTIZATION_METHOD,
            SPARSITY_CONFIG_NAME: {},  # sparsity is removed for now
            TRANSFORM_CONFIG_NAME: tconfig_data,
            **qconfig_data,
        }

        with open(config_file_path, "w") as config_file:
            json.dump(config_data, config_file, indent=2, sort_keys=True)

    def add_decompress_hook(self, model: torch.nn.Module):
        """
        Register a forward pre-hook that decompresses the model on first forward pass.

        The hook automatically removes itself after decompression, allowing the model
        to run in decompressed state for inference.

        :param model: model to attach the decompression hook to
        """

        def ct_decompress_hook(model, args):
            self.decompress_model(model)
            # decompress_model already removes the hook via remove_decompression_hook

        model.ct_decompress_hook = model.register_forward_pre_hook(ct_decompress_hook)

    def remove_decompression_hook(self, model: torch.nn.Module):
        """
        Remove the decompression hook from the model if it exists.

        Called after manual decompression to clean up the hook that would
        otherwise trigger on the next forward pass.

        :param model: model to remove the decompression hook from
        """
        if hasattr(model, "ct_decompress_hook"):
            model.ct_decompress_hook.remove()
            delattr(model, "ct_decompress_hook")



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

# flake8: noqa

from .base import OffloadCache
from .cpu import CPUCache
from .device import DeviceCache
from .disk import DiskCache
from .dist_cpu import DistributedCPUCache
from .dist_device import DistributedDeviceCache
from .dist_disk import DistributedDiskCache



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import contextlib
import inspect
import os
import shutil
from functools import wraps
from types import FrameType

import psutil
import torch
import torch.distributed as dist
from compressed_tensors.offload.convert import from_accelerate
from compressed_tensors.offload.dist_utils import is_distributed, is_rank0
from loguru import logger
from transformers import PreTrainedModel
from transformers.models.auto.modeling_auto import _BaseAutoModelClass


__all__ = ["load_offloaded_model"]


cls_to_patch = _BaseAutoModelClass | PreTrainedModel


@contextlib.contextmanager
def load_offloaded_model(extra_cpu_mem: int = 5e9):
    """
    Context manager used to load a transformers model with offloading implemented by
    compressed-tensors.

    The model is first loaded with accelerate's offloading, then convereted into
    offloading implemented by compressed-tensors. If a distributed environment has been
    initialized, then rank 0 loads the weights while other ranks load on the meta
    device, then the offload is shared across ranks during conversion.

    In addition to the standard `device_map` options, this context also supports
    `device_map="auto_offload"`, which means that the model will load as many parameters
    can fit onto the cpu, and any extra parameters will be loaded on disk.

    :param extra_cpu_mem: extra cpu memory to reserve for any operations not related to
        model loading (bytes). Defaults to 5Gb.
    """
    frame = _get_caller_frame()

    with contextlib.ExitStack() as stack:
        for obj in frame.f_globals.values():
            if isinstance(obj, type) and issubclass(obj, cls_to_patch):
                stack.enter_context(patch_from_pretrained(obj, extra_cpu_mem))

        yield


@contextlib.contextmanager
def patch_from_pretrained(obj: cls_to_patch, extra_cpu_mem: int):
    original_func = obj.from_pretrained.__func__

    @wraps(original_func)
    def from_pretrained(cls, *args, **kwargs):
        kwargs.setdefault("device_map", None)

        # Rank 0 does loading, other ranks init on meta device
        if not is_rank0():
            kwargs["device_map"] = "meta"

        # Intercept `auto_offload`: same as "auto", but only cpu/disk are visible
        elif kwargs["device_map"] == "auto_offload":
            kwargs["device_map"] = "auto"
            if "max_memory" not in kwargs:
                kwargs["max_memory"] = _get_cpu_memory(extra_cpu_mem)

        # Unless the user specifies, use our memory estimates, which take into
        # account distributed setups and extra cpu reserved memory
        elif "max_memory" not in kwargs:
            kwargs["max_memory"] = _get_device_memory() | _get_cpu_memory(extra_cpu_mem)

        model = original_func(cls, *args, **kwargs)
        from_accelerate(model)  # rank 0 shares weights with ranks via offload/broadcast
        return model

    try:
        obj.from_pretrained = from_pretrained.__get__(obj)
        yield
    finally:
        obj.from_pretrained = original_func.__get__(obj)


def _get_device_memory() -> dict[int, int]:
    # TODO: extend to xpu, ect.
    if is_distributed():
        index = dist.get_rank()
        return {index: torch.cuda.get_device_properties(index).total_memory}
    else:
        return {
            index: torch.cuda.get_device_properties(index).total_memory
            for index in range(torch.cuda.device_count())
        }


def _get_cpu_memory(extra_cpu_mem: int) -> dict[str, int]:
    if is_distributed():
        return {"cpu": _get_shared_memory() - extra_cpu_mem}
    else:
        return {"cpu": psutil.virtual_memory().available - extra_cpu_mem}


def _get_shared_memory() -> int:
    linux_shm_path = "/dev/shm"
    if os.path.exists(linux_shm_path):
        total, _used, _free = shutil.disk_usage(linux_shm_path)
        return total

    else:
        logger.warning(
            "Could not find shared memory at `/dev/shm`. Please add platform suppport"
        )
        return psutil.virtual_memory().available


def _get_caller_frame() -> FrameType:
    frame = inspect.currentframe()
    frame = frame.f_back.f_back  # skip this function's caller's frame
    while frame is not None and "contextlib" in frame.f_code.co_filename:
        frame = frame.f_back  # skip contextlib frames

    if frame is None:
        raise RuntimeError("Could not find caller frame")

    return frame



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

"""
Helper functions for packing and unpacking FP4 (E2M1) quantized weights.

FP4 E2M1 format uses 1 sign bit, 2 exponent bits, and 1 mantissa bit,
supporting 8 positive and 8 negative values. This module provides efficient
packing of two FP4 values into a single uint8 for storage.
"""

import torch


__all__ = ["pack_fp4_to_uint8", "unpack_fp4_from_uint8"]


FLOAT_TO_E2M1 = [
    0.0,
    0.5,
    1.0,
    1.5,
    2.0,
    3.0,
    4.0,
    6.0,
]

kE2M1ToFloat = torch.tensor(
    [0.0, 0.5, 1.0, 1.5, 2.0, 3.0, 4.0, 6.0], dtype=torch.float32
)


def pack_fp4_to_uint8(x: torch.Tensor) -> torch.Tensor:
    """
    Packs a tensor with values in the fp4 range into uint8.
    As there are 16 valid fp4 values, two fp4 values can be
    packed into one uint8. Each fp4 value is mapped to its
    particular index (e.g. 0.5 is mapped to index 1, 6.0 is mapped
    to index 7) which is then represented using 4 bits. Consecutive
    pairs of 4 bits are then packed into an uint8.

    :param x: tensor to pack
    :returns: a packed tensor in uint8
    """
    m, n = x.shape
    device = x.device

    if n % 2 != 0:
        raise ValueError(
            "tensor must have an even number of columns for nvfp4 compression"
        )

    # Create lookup table for FP4 values to indices
    kE2M1 = torch.tensor(FLOAT_TO_E2M1, device=device, dtype=x.dtype)

    # Find closest valid FP4 value index for each element
    abs_x = torch.abs(x)
    abs_diff_x = torch.abs(abs_x.unsqueeze(-1) - kE2M1)  # [m, n, 8]
    abs_indices = torch.argmin(abs_diff_x, dim=-1)  # [m, n]

    # Apply sign bit (bit 3) to get final 4-bit representation
    indices = abs_indices + (torch.signbit(x).to(torch.long) << 3)

    # Reshape to prepare for packing pairs of values
    indices = indices.reshape(-1)

    # Reshape to pair consecutive elements
    indices = indices.reshape(-1, 2)

    # Pack pairs of 4-bit values into 8-bit values
    packed = (indices[:, 0] | (indices[:, 1] << 4)).to(torch.uint8)

    return packed.reshape(m, n // 2)


# reference: https://github.com/vllm-project/vllm/pull/16362
def unpack_fp4_from_uint8(
    a: torch.Tensor, m: int, n: int, dtype: torch.dtype | None = torch.bfloat16
) -> torch.Tensor:
    """
    Unpacks uint8 values into fp4. Each uint8 consists of two fp4 values
    (i.e. first four bits correspond to one fp4 value, last four correspond to a
    consecutive fp4 value). The bits represent an index, which are mapped to an fp4
    value.

    :param a: tensor to unpack
    :param m: original dim 0 size of the unpacked tensor
    :param n: original dim 1 size of the unpacked tensor
    :param dtype: dense dtype to cast the unpacked tensor to
    """
    assert a.dtype == torch.uint8

    # Vectorized nibble processing
    a_flat = a.flatten()
    high = (a_flat & 0xF0) >> 4  # Upper nibbles
    low = a_flat & 0x0F  # Lower nibbles

    # Combine nibbles for batch processing
    combined = torch.stack((low, high), dim=1).flatten()

    # Vectorized sign and magnitude extraction
    signs = (combined & 0x08).to(torch.bool)  # Sign bits
    abs_vals = (combined & 0x07).to(torch.long)  # Magnitude indices

    # Device-aware lookup and sign application
    kE2M1 = kE2M1ToFloat.to(device=a.device)
    values = kE2M1[abs_vals] * torch.where(signs, -1.0, 1.0)

    # Reshape to final form
    return values.reshape(m, n).to(dtype=dtype)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

# flake8: noqa
from .base import *
from .dense import *
from .sparse_24_bitmask import *
from .sparse_bitmask import *



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

from enum import Enum

from compressed_tensors.offload.module import unwrap_offload_forward
from torch.nn import Module


__all__ = ["QuantizationMetadata", "KVCacheScaleType"]


class KVCacheScaleType(Enum):
    KEY = "k_scale"
    VALUE = "v_scale"


class QuantizationMetadata:
    """
    Container class for metadata related to quantization
    """

    @staticmethod
    def all_qparam_names():
        """
        All quantization parameter names that might be registered
        onto a module during lifecycle (excluding serialized parameters)
        """
        return [KVCacheScaleType.KEY.value, KVCacheScaleType.VALUE.value] + [
            f"{base_name}_{suffix}"
            for base_name in ("input", "weight", "output")
            for suffix in (
                "global_scale",
                "scale",
                "zero_point",
                "g_idx",
            )
        ]

    @classmethod
    def clear_all_qparams(cls, module: Module):
        """
        Remove all parameters related to quantization that might have
        been registered onto a module previously in lifecycle (excluding
        serialized parameters)

        :param module: Module to clear
        """
        for key in cls.all_qparam_names():
            if hasattr(module, key):
                delattr(module, key)

    @classmethod
    def clear_quantization(cls, module: Module):
        """
        Remove all artifacts of quantization from module, non-recursively.
        Artifacts include any qparams, quantization_scheme, or wrapped
        forward method that might have been altered previously in lifecycle.

        `quantization_status` and `quantization_enabled` are left unchanged.

        :param module: Module to clear
        """
        with unwrap_offload_forward(module):
            # Unwrap forward call
            if hasattr(module.forward, "__wrapped__"):
                module.forward = module.forward.__wrapped__.__get__(module)

            # Clear any qparams
            cls.clear_all_qparams(module)

            # Clear quantization_scheme
            if hasattr(module, "quantization_scheme"):
                delattr(module, "quantization_scheme")



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

from types import SimpleNamespace

from compressed_tensors import ParameterizedDefaultDict, patch_attr, patch_attrs


def test_patch_attr():
    # patch, original value
    obj = SimpleNamespace()
    obj.attribute = "original"
    with patch_attr(obj, "attribute", "patched"):
        assert obj.attribute == "patched"
        obj.attribute = "modified"
    assert obj.attribute == "original"

    # patch, no original attribute
    obj = SimpleNamespace()
    with patch_attr(obj, "attribute", "patched"):
        assert obj.attribute == "patched"
        obj.attribute = "modified"
    assert not hasattr(obj, "attribute")


def test_patch_attrs():
    num_objs = 4
    objs = [SimpleNamespace() for _ in range(num_objs)]
    for idx, obj in enumerate(objs):
        if idx % 2 == 0:
            obj.attribute = f"original_{idx}"
    with patch_attrs(objs, "attribute", [f"patched_{idx}" for idx in range(num_objs)]):
        for idx, obj in enumerate(objs):
            assert obj.attribute == f"patched_{idx}"
            obj.attribute = "modified"
    for idx, obj in enumerate(objs):
        if idx % 2 == 0:
            assert obj.attribute == f"original_{idx}"
        else:
            assert not hasattr(obj, "attribute")


def test_parameterized_default_dict():
    def add_one(value):
        return value + 1

    add_dict = ParameterizedDefaultDict(add_one)
    assert add_dict[0] == 1
    assert add_dict[1] == 2

    def sum_vals(a, b):
        return a + b

    sum_dict = ParameterizedDefaultDict(sum_vals)
    assert sum_dict[0, 1] == 1
    assert sum_dict[5, 7] == 12



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import json

import pytest
from compressed_tensors.config import CompressionFormat
from compressed_tensors.quantization import QuantizationConfig, QuantizationScheme


def test_compression_format_serializable():
    """Test that CompressionFormat can be serialized to JSON"""
    format = CompressionFormat.int_quantized

    # Test direct JSON serialization
    json_str = json.dumps(format)
    assert json_str == '"int-quantized"'

    # Test deserialization
    deserialized = CompressionFormat(json.loads(json_str))
    assert deserialized == format


def test_compression_format_all_values():
    """Test that all CompressionFormat values are serializable"""
    for format in CompressionFormat:
        # Serialize to JSON
        json_str = json.dumps(format)
        assert isinstance(json_str, str)

        # Deserialize from JSON
        deserialized = CompressionFormat(json.loads(json_str))
        assert deserialized == format


def test_compression_format_in_dict():
    """Test that CompressionFormat can be serialized in a dict"""
    test_dict = {
        "format": CompressionFormat.pack_quantized,
        "other_field": "value",
    }

    # Serialize to JSON
    json_str = json.dumps(test_dict, default=str)
    parsed = json.loads(json_str)

    assert parsed["format"] == "pack-quantized"
    assert parsed["other_field"] == "value"


def test_compression_format_in_scheme():
    """Test that CompressionFormat serializes properly in QuantizationScheme"""
    scheme = QuantizationScheme(
        targets=["Linear"], format=CompressionFormat.int_quantized
    )

    # Serialize to dict
    scheme_dict = scheme.model_dump()
    assert scheme_dict["format"] == "int-quantized"
    assert isinstance(scheme_dict["format"], str)

    # Serialize to JSON
    json_str = json.dumps(scheme_dict)
    parsed = json.loads(json_str)
    assert parsed["format"] == "int-quantized"

    # Deserialize from dict
    reloaded = QuantizationScheme.model_validate(parsed)
    assert reloaded.format == CompressionFormat.int_quantized


def test_compression_format_in_config():
    """Test that CompressionFormat serializes properly in QuantizationConfig"""
    config = QuantizationConfig(
        config_groups={"group_1": QuantizationScheme(targets=[])},
        format=CompressionFormat.float_quantized.value,
    )

    # Serialize to dict
    config_dict = config.to_dict()
    assert config_dict["format"] == "float-quantized"
    assert isinstance(config_dict["format"], str)

    # Serialize to JSON
    json_str = json.dumps(config_dict)
    parsed = json.loads(json_str)
    assert parsed["format"] == "float-quantized"

    # Deserialize from dict
    reloaded = QuantizationConfig.model_validate(parsed)
    assert reloaded.format == "float-quantized"


@pytest.mark.parametrize(
    "format_value",
    [
        CompressionFormat.dense,
        CompressionFormat.int_quantized,
        CompressionFormat.float_quantized,
        CompressionFormat.pack_quantized,
        CompressionFormat.naive_quantized,
        CompressionFormat.mixed_precision,
        CompressionFormat.nvfp4_pack_quantized,
        CompressionFormat.mxfp4_pack_quantized,
    ],
)
def test_compression_format_round_trip(format_value):
    """Test round-trip serialization for each CompressionFormat value"""
    # Create a config with the format
    config = QuantizationConfig(
        config_groups={"group_1": QuantizationScheme(targets=["Linear"])},
        format=format_value.value,
    )

    # Serialize to dict then JSON
    config_dict = config.to_dict()
    json_str = json.dumps(config_dict)

    # Deserialize from JSON
    parsed = json.loads(json_str)
    reloaded = QuantizationConfig.model_validate(parsed)

    # Verify format is preserved
    assert reloaded.format == format_value.value
    assert config == reloaded



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

"""
Helper functions for packing and unpacking quantized weights into int32 format.

These functions enable efficient storage of sub-8-bit quantized weights by packing
multiple values into 32-bit integers.
"""

import math
from typing import Literal

import torch


__all__ = ["pack_to_int32", "unpack_from_int32"]


def pack_to_int32(
    value: torch.Tensor,
    num_bits: int,
    packed_dim: Literal[0, 1] = 1,
) -> torch.Tensor:
    """
    Packs a tensor of quantized weights stored in int8 into int32s with padding

    Pseudocode:
     1. Shift wrt num_bits to convert to unsigned. num_bits=8
        [1,2] -> [129, 130]
     2. Pad to fill in 32 bits
        [129, 130] -> [129, 130, 0, 0]
     3. convert to binary align in order
        [129, 130, 0, 0] -> 00000000 00000000 10000010 10000001
     4. convert aligned binary to number
        00000000000000001000001010000001 -> 33409
     5. covert back to uint32
        33409 -> 33409

    :param value: tensor to pack
    :param num_bits: number of bits used to store underlying data, must be at least 1
    :returns: packed int32 tensor
    """
    if value.dtype is not torch.int8:
        raise ValueError("Tensor must be quantized to torch.int8 before packing")

    if num_bits > 8:
        raise ValueError("Packing is only supported for less than 8 bits")

    if num_bits < 1:
        raise ValueError(f"num_bits must be at least 1, got {num_bits}")

    # Handle N-dimensional tensors (e.g. MoE 3D weights) by packing each 2D slice
    if value.ndim > 2:
        return torch.stack(
            [
                pack_to_int32(value[i], num_bits, packed_dim)
                for i in range(value.shape[0])
            ]
        )

    # Convert to unsigned range for packing, matching quantization offset
    offset = 1 << (num_bits - 1)
    value = (value + offset).to(torch.uint8)
    device = value.device

    pack_factor = 32 // num_bits

    if packed_dim == 0:
        value = value.transpose(0, 1)

    rows, cols = value.shape
    padded_cols = math.ceil(cols / pack_factor) * pack_factor
    pad_len = padded_cols - cols

    if pad_len > 0:
        value = torch.nn.functional.pad(value, (0, pad_len))

    num_groups = padded_cols // pack_factor

    # Use int32 here
    reshaped = value.view(rows, num_groups, pack_factor).to(torch.int32)
    bit_shifts = torch.arange(pack_factor, device=device, dtype=torch.int32) * num_bits
    packed = (reshaped << bit_shifts).sum(dim=2, dtype=torch.int32)

    if packed_dim == 0:
        packed = packed.transpose(0, 1)

    return packed


def unpack_from_int32(
    value: torch.Tensor,
    num_bits: int,
    shape: torch.Size,
    packed_dim: Literal[0, 1] = 1,
) -> torch.Tensor:
    """
    Unpacks a tensor of packed int32 weights into individual int8s, maintaining the
    original bit range.

    Return tensors in int8

    :param value: tensor to upack
    :param num_bits: number of bits to unpack each data point into
    :param shape: shape to unpack into, used to remove padding
    :returns: unpacked int8 tensor
    """
    if value.dtype is not torch.int32:
        raise ValueError(
            f"Expected {torch.int32} but got {value.dtype}, Aborting unpack."
        )

    if num_bits > 8:
        raise ValueError("Unpacking is only supported for less than 8 bits")

    # Handle N-dimensional tensors (e.g. MoE 3D weights) by unpacking each 2D slice
    if value.ndim > 2:
        return torch.stack(
            [
                unpack_from_int32(value[i], num_bits, shape[1:], packed_dim)
                for i in range(value.shape[0])
            ]
        )

    pack_factor = 32 // num_bits

    # unpack
    mask = (1 << num_bits) - 1

    if packed_dim == 1:
        unpacked = torch.zeros(
            (value.shape[0], value.shape[1] * pack_factor),
            device=value.device,
            dtype=torch.int32,
        )
        for i in range(pack_factor):
            unpacked[:, i::pack_factor] = (value >> (num_bits * i)) & mask

        # remove padding
        original_row_size = int(shape[1])
        unpacked = unpacked[:, :original_row_size]
    else:
        unpacked = torch.zeros(
            (value.shape[0] * pack_factor, value.shape[1]),
            device=value.device,
            dtype=torch.int32,
        )
        for i in range(pack_factor):
            unpacked[i::pack_factor, :] = (value >> (num_bits * i)) & mask

        # remove padding
        original_row_size = int(shape[0])
        unpacked = unpacked[:original_row_size, :]

    # bits are packed in unsigned format, reformat to signed
    # update the value range from unsigned to signed
    offset = pow(2, num_bits) // 2
    unpacked = (unpacked - offset).to(torch.int8)

    return unpacked



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import os

import pytest
import torch
from compressed_tensors.offload import dispatch_with_map, offload_module, to_accelerate
from compressed_tensors.offload.convert.to_accelerate import to_accelerate_module
from tests.test_offload.conftest import torchrun
from tests.testing_utils import requires_gpu


acclerate = pytest.importorskip("accelerate")


@pytest.mark.unit
@requires_gpu
@pytest.mark.parametrize("offload_device", ["cuda", "cuda:0", "cpu", "disk"])
def test_to_accelerate_module(offload_device, tmp_path):
    linear = torch.nn.Linear(5, 5)
    if offload_device == "disk":
        offload_dir = tmp_path / "offload_dir"
        os.mkdir(offload_dir)
        offload_module(linear, "cuda", offload_device, offload_dir=str(offload_dir))
    else:
        offload_module(linear, "cuda", offload_device)

    _offload_device = to_accelerate_module(linear, name="", hf_disk_index={})
    if offload_device == "cuda":
        assert _offload_device == "cuda:0"
    else:
        assert _offload_device == offload_device


@pytest.mark.unit
@requires_gpu
def test_to_accelerate(cuda_device, tmp_path):
    offload_dir = tmp_path / "offload_dir"
    os.mkdir(offload_dir)

    model = torch.nn.Sequential(
        torch.nn.Linear(5, 5), torch.nn.Linear(5, 5), torch.nn.Linear(5, 5)
    )
    device_map = {
        "0": (torch.device("cuda"), torch.device("cpu")),
        "1": (torch.device("cuda"), torch.device("cuda")),
        "2": (torch.device("cuda"), "disk"),
    }
    dispatch_with_map(model, device_map, offload_dir)

    hf_device_map = to_accelerate(model)
    assert hf_device_map == {"": "cpu", "0": "cpu", "1": str(cuda_device), "2": "disk"}
    assert hasattr(model[0], "_hf_hook")
    assert hasattr(model[1], "_hf_hook")
    assert hasattr(model[2], "_hf_hook")


@pytest.mark.unit
@requires_gpu(2)
@torchrun(world_size=2)
def test_to_accelerate_dist(cuda_device, tmp_path):
    test_to_accelerate(cuda_device, tmp_path)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import os
from typing import TYPE_CHECKING, Literal

import torch
import torch.distributed as dist
from compressed_tensors.offload.cache import DiskCache
from compressed_tensors.offload.convert.helpers import (
    DEFAULT_OFFLOAD_DEVICE,
    get_tensors,
    norm_device,
)
from compressed_tensors.offload.dispatch import dispatch_with_map
from compressed_tensors.offload.dist_utils import is_distributed, is_rank0
from compressed_tensors.offload.utils import to_tensor
from loguru import logger


if TYPE_CHECKING:
    from accelerate.utils import OffloadedWeightsLoader
    from compressed_tensors.offload.dispatch import DeviceMap


__all__ = ["from_accelerate", "remove_accelerate", "remove_accelerate_from_module"]


def from_accelerate(model: torch.nn.Module) -> tuple["DeviceMap", str | None]:
    """
    Convert a model from accelerate offloading to compressed-tensors offloading. Often
    called by `load_offloaded_model` to load offloaded models across ranks.

    Rank 0 is always expected to provide an accelerate-offloaded model.
    Other ranks (if they exist) may provide a model on any device, with or
    without accelerate offloading.

    - If called after `load_offloaded_model()`, other ranks will provide a meta model
        with no accelerate offloading
    - If called after `to_accelerate`, other ranks will provide an accelerate-offloaded
        model shared cpu tensors/file paths.

    :param model: accelerate-offloaded model if rank0, meta model otherwise
    """
    device_map, offload_dir = remove_accelerate(model)

    broadcast_obj = [device_map, offload_dir]
    if is_distributed():
        dist.broadcast_object_list(broadcast_obj, src=0)

    dispatch_with_map(model, *broadcast_obj)
    return tuple(broadcast_obj)


def remove_accelerate(model: torch.nn.Module) -> tuple["DeviceMap", str | None]:
    """
    Remove accelerate offloading from a model, if applicable

    :param model: model containing accelerate offloaded modules
    :returns: `(device_map, offload_dir)`
    """
    offload_dir = None
    device_map = {}

    for name, module in model.named_modules(remove_duplicate=False):
        onload_dev, offload_dev, _offload_dir = remove_accelerate_from_module(module)

        if _offload_dir is not None:
            if offload_dir is not None and _offload_dir != offload_dir:
                raise ValueError(
                    "Expected model to only have one `offload_dir`, "
                    f"instead got {offload_dir} and {_offload_dir}"
                )

            offload_dir = _offload_dir

        device_map[name] = (onload_dev, offload_dev)

    if hasattr(model, "hf_device_map"):
        delattr(model, "hf_device_map")

    return device_map, offload_dir


def remove_accelerate_from_module(
    module: torch.nn.Module,
) -> tuple[torch.device | None, torch.device | Literal["disk"] | None, str | None]:
    """
    Remove accelerate offloading from a module, if present.
    Absolutely no device movement occurs, and parameters/buffers pointers from state
    dicts coming from `to_accelerate` remain unchanged so as to avoid memory duplication

    :param module: module to remove offloading from
    :returns: `(onload_device, offload_device, disk_offload_dir)`
    """
    try:
        from accelerate.hooks import AlignDevicesHook, remove_hook_from_module
        from accelerate.utils import OffloadedWeightsLoader, PrefixedDataset
    except ImportError:
        device = _infer_module_device(module)
        return device, device, None

    hook = getattr(module, "_hf_hook", None)
    direct_tensors = _direct_tensors(module)

    # No AlignDevicesHook: treat as "not offloaded"
    if not isinstance(hook, AlignDevicesHook):
        device = _infer_device_from_tensors(direct_tensors)
        return device, device, None

    # Hook exists but no active offload (or nothing to consider)
    if not hook.offload or not direct_tensors:
        hook.offload = False
        remove_hook_from_module(module, recurse=False)
        device = _infer_device_from_tensors(direct_tensors)
        return device, device, None

    # Unwrap PrefixedDataset chain so we can look up real tensor keys
    prefix, dataset = _unwrap_prefixed_dataset(hook.weights_map, PrefixedDataset)
    assert isinstance(dataset, (OffloadedWeightsLoader, dict))

    offload_dev: str | None = None

    for local_name, tensor in direct_tensors.items():
        full_name = prefix + local_name

        # Device/CPU offload
        if isinstance(dataset, dict) and local_name in dataset:
            offload = dataset[local_name]
            offload_dev = _set_or_validate_offload(offload_dev, offload.device.type)

        # Device/CPU offload
        elif (
            isinstance(dataset, OffloadedWeightsLoader)
            and full_name in dataset.state_dict
        ):
            offload = dataset.state_dict[full_name]
            offload_dev = _set_or_validate_offload(offload_dev, offload.device.type)

        # Disk offload
        elif isinstance(dataset, OffloadedWeightsLoader) and full_name in dataset.index:
            offload = tensor
            offload_dev = _set_or_validate_offload(offload_dev, "disk")
            assert offload.device.type == "meta"
            assert isinstance(offload, (torch.nn.Parameter, torch.nn.Buffer))

            # Copy accelerate's disk index into DiskCache for our later use
            if is_rank0():
                _save_ct_index_entry(dataset, full_name, tensor)

        # Not offloaded, likely a buffer
        else:
            offload = tensor

        # Replace meta tensor with offloaded value (no ptr rematerialization occurs)
        # In the disk case, the tensor remains as the meta tensor
        if not isinstance(offload, (torch.nn.Parameter, torch.nn.Buffer)):
            to_tensor(offload, tensor)
        setattr(module, local_name, offload)

    # Prevent onloading disk tensors while removing hook
    hook.offload = False
    remove_hook_from_module(module, recurse=False)

    return (
        norm_device(hook.execution_device),
        norm_device(offload_dev if offload_dev is not None else DEFAULT_OFFLOAD_DEVICE),
        dataset.save_folder if isinstance(dataset, OffloadedWeightsLoader) else None,
    )


def _save_ct_index_entry(
    dataset: "OffloadedWeightsLoader",
    name: str,
    offloaded: torch.Tensor,
):
    entry: dict = dataset.index[name]

    if "safetensors_file" in entry and offloaded not in DiskCache.index:
        # typical case: model is loaded from safetensors file
        # create a symlink that points to the model safetensor file
        # if the value is ever updated, the symlink is broken and a real file
        # is written to that location
        DiskCache.create_checkpoint_symlink(offloaded, entry, dataset.save_folder)

    else:
        # unfortunately, ct's implementation does not support loading non-safetensors
        # we must onload and save as safetensors. This typically only occurs in testing
        onloaded = dataset[name]
        DiskCache("cpu", offload_dir=dataset.save_folder).offload(
            onloaded, offloaded=offloaded
        )
        logger.warning(
            "Attempting to disk offload a model which was not saved with safetensors. "
            "compressed-tensors only supports disk onload from safetensors files, so "
            "weights must be onloaded and re-saved as safetensors files.",
            log_once=True,
        )

        # remove original weight_file
        original_weight_file = os.path.join(dataset.save_folder, f"{name}.dat")
        if os.path.exists(original_weight_file):
            os.remove(original_weight_file)


def _direct_tensors(module: torch.nn.Module) -> dict[str, torch.Tensor]:
    return {name: t for name, t in get_tensors(module) if t is not None}


def _infer_device_from_tensors(tensors: dict[str, torch.Tensor]) -> torch.device | None:
    t = next(iter(tensors.values()), None)
    return norm_device(t.device if t is not None else None)


def _infer_module_device(module: torch.nn.Module) -> torch.device | None:
    return _infer_device_from_tensors(_direct_tensors(module))


def _unwrap_prefixed_dataset(weights_map, PrefixedDatasetType):
    prefix = ""
    dataset = weights_map
    while isinstance(dataset, PrefixedDatasetType):
        prefix += dataset.prefix
        dataset = dataset.dataset
    return prefix, dataset


def _set_or_validate_offload(current: str | None, new: str) -> str:
    if current not in (None, new):
        raise ValueError("Expected all accelerate tensors to share offload")
    return new



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import os

from compressed_tensors.entrypoints.convert import Converter
from safetensors.torch import load_file, save_file


__all__ = [
    "validate_file",
    "convert_file",
]


def validate_file(
    file_path: str | os.PathLike,
    converter: Converter,
):
    """
    Validate that each quantizable tensor in a safetensors file can be quantized.

    :param file_path: safetensors file to validate
    :param converter: converter we wish to apply to the checkpoint,
        e.g. conversion of some layers from some format to compressed-tensors
    """
    tensors = load_file(file_path)

    converter.validate(tensors)


def convert_file(
    file_path: str | os.PathLike,
    save_path: str | os.PathLike,
    converter: Converter,
) -> tuple[int, dict[str, str]]:
    """
    Convert tensors in a given safetensors file

    :param file_path: safetensors file to process
    :param save_path: save path of file with quantized weights
    :param converter: converter we wish to apply to the checkpoint,
        e.g. conversion of some layers from some format to compressed-tensors
    """
    tensors = load_file(file_path)

    converter.process(tensors)

    save_file(tensors, save_path)
    total_size = sum(tensor.nbytes for tensor in tensors.values())
    weight_map = {key: os.path.basename(save_path) for key in tensors.keys()}
    return total_size, weight_map



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import torch
from compressed_tensors.compressors.base import BaseCompressor
from compressed_tensors.config import CompressionFormat
from compressed_tensors.quantization import (
    QuantizationScheme,
    QuantizationStrategy,
    QuantizationType,
)
from compressed_tensors.quantization.lifecycle.forward import dequantize, quantize
from compressed_tensors.quantization.utils import maybe_pad_tensor_for_block_quant
from compressed_tensors.utils import TensorStateDict


__all__ = [
    "NaiveQuantizationCompressor",
    "IntQuantizationCompressor",
    "FloatQuantizationCompressor",
]


@BaseCompressor.register(name=CompressionFormat.naive_quantized.value)
class NaiveQuantizationCompressor(BaseCompressor):
    """
    Naive quantization compressor.

    Each quantized layer's weight is converted from its original float dtype to
    the closest PyTorch dtype for the bit-width specified by QuantizationArgs.
    """

    @classmethod
    def compress(
        cls, state_dict: TensorStateDict, scheme: QuantizationScheme
    ) -> TensorStateDict:
        """
        Compress a per-module state dict.

        Quantizes the weight to the dtype specified by the scheme's
        QuantizationArgs. Handles block quantization padding if needed.

        :param state_dict: local-name state dict (weight, weight_scale, …)
        :param scheme: quantization scheme for the weight
        :return: compressed state dict
        """
        state_dict = state_dict.copy()
        weight = state_dict.pop("weight")
        scale = state_dict.get("weight_scale")
        zero_point = state_dict.get("weight_zero_point", None)
        g_idx = state_dict.get("weight_g_idx", None)
        weights = scheme.weights

        original_weight_shape = weight.shape

        # For block quantization, pad weight to divisible dimensions
        if (
            weights.strategy == QuantizationStrategy.BLOCK
            and weights.block_structure is not None
        ):
            block_structure = tuple(weights.block_structure)
            weight = maybe_pad_tensor_for_block_quant(weight, block_structure)

        quantized_weight = quantize(
            x=weight,
            scale=scale,
            zero_point=zero_point,
            g_idx=g_idx,
            args=weights,
            dtype=weights.pytorch_dtype(),
        )

        # Truncate back to original shape if padding was added
        if quantized_weight.shape != original_weight_shape:
            quantized_weight = quantized_weight[
                tuple([slice(v) for v in original_weight_shape])
            ]

        state_dict["weight"] = quantized_weight
        state_dict = cls._remove_symmetric_zp(state_dict, scheme)

        return state_dict

    @classmethod
    def decompress(
        cls, state_dict: TensorStateDict, scheme: QuantizationScheme
    ) -> TensorStateDict:
        """
        Decompress a per-module state dict.

        Dequantizes the weight back to float dtype using the scale and
        zero-point from the state dict.

        :param state_dict: local-name state dict (weight, weight_scale, …)
        :param scheme: quantization scheme for the weight
        :return: decompressed state dict with weight in float dtype
        """
        state_dict = state_dict.copy()
        weight = state_dict.pop("weight")
        scale = state_dict.get("weight_scale")
        zero_point = state_dict.get("weight_zero_point", None)
        g_idx = state_dict.get("weight_g_idx", None)

        state_dict["weight"] = dequantize(
            x_q=weight,
            scale=scale,
            zero_point=zero_point,
            g_idx=g_idx,
        )

        return state_dict

    @classmethod
    def can_compress(cls, module_type: type, scheme: QuantizationScheme) -> bool:
        """
        Naive quantization is the fallback compressor - it matches any quantized
        scheme that doesn't match a more specific compressor.
        """
        return module_type == torch.nn.Linear and scheme.weights is not None


@BaseCompressor.register(name=CompressionFormat.int_quantized.value)
class IntQuantizationCompressor(NaiveQuantizationCompressor):
    """Alias for integer quantized models."""

    @classmethod
    def can_compress(cls, module_type: type, scheme: QuantizationScheme) -> bool:
        """Int quantized matches w8a8 int quantization."""
        return (
            module_type == torch.nn.Linear
            and scheme.input_activations is not None
            and scheme.weights is not None
            and scheme.weights.type == QuantizationType.INT.value
        )


@BaseCompressor.register(name=CompressionFormat.float_quantized.value)
class FloatQuantizationCompressor(NaiveQuantizationCompressor):
    """Alias for fp quantized models."""

    @classmethod
    def can_compress(cls, module_type: type, scheme: QuantizationScheme) -> bool:
        """Float quantized matches w8a8 float quantization."""
        return (
            module_type == torch.nn.Linear
            and scheme.input_activations is not None
            and scheme.weights is not None
            and scheme.weights.type == QuantizationType.FLOAT.value
        )



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import pytest
import torch
from compressed_tensors import IntQuantizationCompressor
from compressed_tensors.quantization import (
    QuantizationArgs,
    QuantizationScheme,
    QuantizationStrategy,
)
from compressed_tensors.quantization.lifecycle.forward import fake_quantize


def make_quant_scheme(strategy, group_size=None, symmetric=True):
    return QuantizationScheme(
        targets=["Linear"],
        weights=QuantizationArgs(
            strategy=strategy, group_size=group_size, symmetric=symmetric
        ),
    )


@pytest.mark.parametrize(
    "strategy,symmetric,group_size,sc,zp",
    [
        [QuantizationStrategy.TENSOR, True, None, torch.tensor(0.01), torch.tensor(0)],
        [
            QuantizationStrategy.GROUP,
            True,
            128,
            torch.rand((512, 8)) * 0.01,
            torch.zeros((512, 8), dtype=torch.int8),
        ],
        [
            QuantizationStrategy.CHANNEL,
            False,
            None,
            torch.rand((512, 1)) * 0.01,
            ((torch.rand((512, 1)) - 0.5) * 127).to(torch.int8),
        ],
    ],
)
def test_quant_format(strategy, symmetric, group_size, sc, zp):
    module_sd = {
        "weight": torch.rand((512, 1024)),
        "weight_scale": sc.to(torch.float32),
        "weight_zero_point": zp.to(torch.int32),
    }
    scheme = make_quant_scheme(
        strategy=strategy, group_size=group_size, symmetric=symmetric
    )

    compressed = IntQuantizationCompressor.compress(module_sd, scheme=scheme)

    # zero_point is dropped for symmetric quantization
    if symmetric:
        assert "weight_zero_point" not in compressed
    else:
        assert compressed["weight_zero_point"].dtype == torch.int32

    # weight should be compressed to int8
    assert compressed["weight"].dtype == torch.int8
    assert compressed["weight_scale"].dtype == torch.float32


@pytest.mark.parametrize(
    "strategy,group_size,sc,zp",
    [
        [QuantizationStrategy.TENSOR, None, torch.tensor(0.01), torch.tensor(0)],
        [
            QuantizationStrategy.GROUP,
            128,
            torch.rand((300, 8)) * 0.01,
            torch.zeros((300, 8), dtype=torch.int8),
        ],
        [
            QuantizationStrategy.CHANNEL,
            None,
            torch.rand((300, 1)) * 0.01,
            torch.zeros((300, 1), dtype=torch.int8),
        ],
    ],
)
def test_compress_decompress_match(strategy, group_size, sc, zp):
    module_sd = {
        "weight": torch.rand((300, 1024)),
        "weight_scale": sc.to(torch.float32),
        "weight_zero_point": zp.to(torch.int32),
    }
    scheme = make_quant_scheme(strategy=strategy, group_size=group_size)

    compressed = IntQuantizationCompressor.compress(module_sd, scheme=scheme)
    decompressed = IntQuantizationCompressor.decompress(compressed, scheme=scheme)

    fake_quant = fake_quantize(
        module_sd["weight"],
        scale=module_sd["weight_scale"],
        zero_point=module_sd["weight_zero_point"],
        args=scheme.weights,
    )
    assert torch.equal(fake_quant, decompressed["weight"].to(torch.float32))



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

# flake8: noqa

from .helpers import *
from .internal import *
from .match import *
from .module import *
from .offload import *
from .permutations_24 import *
from .safetensors_load import *
from .semi_structured_conversions import *
from .type import *



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import os
import subprocess
import sys
from functools import wraps
from types import FunctionType
from typing import Any, Callable, Literal, Optional

import pytest
import torch
import torch.distributed as dist
from compressed_tensors.offload.utils import send_tensors


def assert_device_equal(
    device_a: torch.device | Literal["disk"],
    device_b: torch.device | Literal["disk"],
):
    if device_a == "disk":
        device_a = torch.device("meta")
    if device_b == "disk":
        device_b = torch.device("meta")

    cur_index = torch.cuda.current_device()
    a_index = cur_index if device_a.index is None else device_a.index
    b_index = cur_index if device_b.index is None else device_b.index
    assert device_a.type == device_b.type and a_index == b_index


def assert_tensor_equal(
    tensor_a: torch.Tensor,
    tensor_b: torch.Tensor,
    device: Optional[torch.device | str] = None,
):
    if device is not None:
        tensor_b = send_tensors(tensor_b, "meta" if device == "disk" else device)

    assert tensor_a.__class__ == tensor_b.__class__
    assert tensor_a.requires_grad == tensor_b.requires_grad
    assert tensor_a.__dict__ == tensor_b.__dict__

    if tensor_a.is_meta or tensor_b.is_meta:
        assert (
            tensor_a.device == tensor_b.device
            and tensor_a.shape == tensor_b.shape
            and tensor_a.dtype == tensor_b.dtype
        )
    else:
        assert torch.equal(tensor_a, tensor_b)


def torchrun(world_size: int = 1) -> Callable[[Callable[..., Any]], Callable[..., Any]]:
    """
    Test a function within parallel `torchrun` subprocesses, each running with `pytest`
    ```
    # (main) -> (torchrun) -
    #              \\-- (rank 0)
    #              \\-- (rank 1)
    ```

    :param world_size: number of ranks to spawn
    """

    def decorator(func: FunctionType):
        @wraps(func)
        def wrapper(*args, **kwargs):
            # We're running in a torchrun subprocess:
            # init distributed and run test func
            if "TORCHELASTIC_RUN_ID" in os.environ:
                rank = int(os.environ["RANK"])
                local_rank = int(os.environ["LOCAL_RANK"])

                torch.cuda.set_device(local_rank)
                dist.init_process_group(
                    backend="nccl",
                    init_method="env://",
                    rank=rank,
                    world_size=world_size,
                    device_id=local_rank,
                )
                dist.barrier()

                return func(*args, **kwargs)

            # First time calling in the main process:
            # trigger torchrun with this function as the pytest target
            else:
                file_path = sys.modules.get(func.__module__).__file__
                func_name = func.__name__

                cmd = (
                    f"{sys.executable} "
                    f"-m torch.distributed.run --nproc_per_node {world_size} "
                    "--log-dir /tmp/torchrun-logs --tee 3 --role torchrun "
                    f"-m pytest {file_path}::{func_name} -sx"
                )

                proc = subprocess.run(cmd.split(" "))
                assert proc.returncode == 0

        return wrapper

    return decorator


@pytest.fixture()
def cuda_device():
    return (
        torch.device("cuda")
        if "TORCHELASTIC_RUN_ID" in os.environ
        else torch.device("cuda:0")
    )



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import torch
from compressed_tensors.compressors.base import BaseCompressor
from compressed_tensors.compressors.mx_utils import (
    compress_mx_scale,
    decompress_mx_scale,
)
from compressed_tensors.compressors.naive_quantized.base import (
    NaiveQuantizationCompressor,
)
from compressed_tensors.config import CompressionFormat
from compressed_tensors.quantization import (
    QuantizationArgs,
    QuantizationScheme,
    QuantizationType,
)
from compressed_tensors.utils import TensorStateDict


__all__ = ["MXFP8QuantizationCompressor"]


@BaseCompressor.register(name=CompressionFormat.mxfp8_quantized.value)
class MXFP8QuantizationCompressor(NaiveQuantizationCompressor):
    """
    Compressor for MXFP8 quantized models.

    Weights are stored as float8_e4m3fn with E8M0 (power-of-2) scales
    stored as uint8 exponents. Extends NaiveQuantizationCompressor by
    converting float scales to/from E8M0 exponent format during
    compression/decompression.
    """

    @classmethod
    def _compress_scale(
        cls, scale: torch.Tensor, weights: QuantizationArgs
    ) -> torch.Tensor:
        assert weights.scale_dtype is not None
        return compress_mx_scale(scale, weights.scale_dtype)

    @classmethod
    def _decompress_scale(cls, scale: torch.Tensor) -> torch.Tensor:
        return decompress_mx_scale(scale)

    @classmethod
    def compress(
        cls, state_dict: TensorStateDict, scheme: QuantizationScheme
    ) -> TensorStateDict:
        """
        Compress a per-module state dict for MXFP8 format.

        Quantizes the weight via the parent class, then converts the
        float scale to E8M0 exponent format (uint8) for storage.

        :param state_dict: local-name state dict (weight, weight_scale, ...)
        :param scheme: quantization scheme for the weight
        :return: compressed state dict with E8M0 scales
        """
        state_dict = NaiveQuantizationCompressor.compress(state_dict, scheme)

        # Convert float scale to E8M0 exponent format (uint8) for storage
        scale = state_dict["weight_scale"]
        state_dict["weight_scale"] = cls._compress_scale(scale, scheme.weights)

        return state_dict

    @classmethod
    def decompress(
        cls, state_dict: TensorStateDict, scheme: QuantizationScheme
    ) -> TensorStateDict:
        """
        Decompress a per-module state dict for MXFP8 format.

        Converts E8M0 exponent scales back to float, then dequantizes
        the weight via the parent class.

        :param state_dict: local-name state dict (weight, weight_scale, ...)
        :param scheme: quantization scheme for the weight
        :return: decompressed state dict with weight in float dtype
        """
        state_dict = state_dict.copy()

        # Convert E8M0 scale back to bfloat16 for consistency with model dtype
        scale = state_dict["weight_scale"]
        state_dict["weight_scale"] = cls._decompress_scale(scale)

        return NaiveQuantizationCompressor.decompress(state_dict, scheme)

    @classmethod
    def can_compress(cls, module_type: type, scheme: QuantizationScheme) -> bool:
        """MXFP8 matches FP8 with group_size=32 and uint8 scale_dtype."""
        return (
            module_type == torch.nn.Linear
            and scheme.weights is not None
            and scheme.weights.num_bits == 8
            and scheme.weights.type == QuantizationType.FLOAT.value
            and scheme.weights.group_size == 32
            and scheme.weights.scale_dtype == torch.uint8
        )



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

from compressed_tensors.modeling import (
    IMPL_ATTR,
    KV_CACHE_ATTR,
    QuantizedAttentionImpl,
    QuantizedKVCache,
    register_query_hook,
)
from compressed_tensors.quantization.lifecycle.apply import apply_quantization_config
from compressed_tensors.quantization.quant_args import (
    QuantizationArgs,
    QuantizationStrategy,
)
from compressed_tensors.quantization.quant_config import (
    QuantizationConfig,
    QuantizationStatus,
)
from compressed_tensors.quantization.quant_scheme import QuantizationScheme
from tests.testing_utils import requires_gpu
from transformers import AutoModelForCausalLM


@requires_gpu
def test_apply_config_detects_deepseekv3_attention_and_hooks():
    model = AutoModelForCausalLM.from_pretrained(
        "trl-internal-testing/tiny-DeepseekV3ForCausalLM", device_map="cuda"
    )
    inputs = {key: value.to("cuda") for key, value in model.dummy_inputs.items()}

    # Build attention quantization scheme targeting attention modules
    qa = QuantizationArgs(
        strategy=QuantizationStrategy.TENSOR,
        symmetric=True,
        num_bits=8,
        dynamic=False,
    )
    scheme = QuantizationScheme(
        targets=["re:.*self_attn$"],
        input_activations=qa,
    )
    config = QuantizationConfig(
        config_groups={"group_0": scheme},
        quantization_status=QuantizationStatus.INITIALIZED,
        kv_cache_scheme=None,
    )
    apply_quantization_config(model, config)

    # Validate q/k/v qparams initialized and hooks attached
    q_called = []
    k_called = []
    v_called = []

    for idx, layer in enumerate(model.model.layers):
        attn = layer.self_attn
        assert isinstance(getattr(attn, IMPL_ATTR), QuantizedAttentionImpl)
        assert isinstance(getattr(attn, KV_CACHE_ATTR), QuantizedKVCache)

        assert hasattr(attn, "q_scale")
        assert hasattr(attn, "k_scale")
        assert hasattr(attn, "v_scale")
        assert hasattr(attn, "q_zero_point")
        assert hasattr(attn, "k_zero_point")
        assert hasattr(attn, "v_zero_point")

        q_called.append(False)
        k_called.append(False)
        v_called.append(False)

        def q_hook(_module, _states, i=idx):
            q_called[i] = True

        register_query_hook(attn, q_hook)
        impl = getattr(attn, IMPL_ATTR)

        def _k_pre_hook(_impl, args, kwargs, i=idx):
            k_called[i] = True
            return args, kwargs

        def _v_pre_hook(_impl, args, kwargs, i=idx):
            v_called[i] = True
            return args, kwargs

        impl.register_forward_pre_hook(_k_pre_hook, with_kwargs=True)
        impl.register_forward_pre_hook(_v_pre_hook, with_kwargs=True)

    model(**inputs, use_cache=True)
    assert all(q_called) and all(k_called) and all(v_called)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

"""
Shared utilities for MX-format (MXFP4/MXFP8) scale compression and
decompression.

MX scales are stored as E8M0 exponents (uint8), where each value represents
a biased power-of-2 exponent with bias 127: ``scale_float = 2 ** (exp - 127)``.
"""

import torch


__all__ = ["compress_mx_scale", "decompress_mx_scale"]


def compress_mx_scale(scale: torch.Tensor, scale_dtype: torch.dtype) -> torch.Tensor:
    """
    Convert a float scale tensor to E8M0 exponent format for MX storage.

    Extracts the power-of-2 exponent from each scale value and adds the
    E8M0 bias (127).

    :param scale: float scale tensor (e.g. from observer output)
    :param scale_dtype: target dtype for the compressed scale (typically uint8)
    :return: biased exponent tensor in the requested dtype
    """
    scale_exp = 127 + torch.floor(torch.log2(scale)).to(torch.int32)
    return scale_exp.to(scale_dtype)


def decompress_mx_scale(scale: torch.Tensor) -> torch.Tensor:
    """
    Convert E8M0 exponent scale back to float (bfloat16).

    Reverses the E8M0 encoding by subtracting the bias and raising 2 to
    the resulting power.

    :param scale: uint8 tensor of biased E8M0 exponents
    :return: bfloat16 tensor of float scale values
    """
    scale_exp = scale.to(torch.int32) - 127
    return 2.0 ** scale_exp.to(torch.bfloat16)



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

# flake8: noqa
# isort: off

from .converters import *
from .file_utils import *
from .save_utils import *
from .convert_checkpoint import *



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

"""
Logger configuration for Compressed Tensors.
"""

import os
import sys
from dataclasses import dataclass
from typing import Any, Dict, Optional

from loguru import logger


__all__ = ["LoggerConfig", "configure_logger", "logger"]


# used by `support_log_once``
_logged_once = set()


@dataclass
class LoggerConfig:
    disabled: bool = False
    clear_loggers: bool = True
    console_log_level: Optional[str] = "INFO"
    log_file: Optional[str] = None
    log_file_level: Optional[str] = None


def configure_logger(config: Optional[LoggerConfig] = None):
    """
    Configure the logger for Compressed Tensors.
    This function sets up the console and file logging
    as per the specified or default parameters.

    Note: Environment variables take precedence over the function parameters.

    :param config: The configuration for the logger to use.
    :type config: LoggerConfig
    """
    logger_config = config or LoggerConfig()

    # env vars get priority
    if bool(os.getenv("COMPRESSED_TENSORS_LOG_DISABLED")):
        logger_config.disabled = True
    if bool(os.getenv("COMPRESSED_TENSORS_CLEAR_LOGGERS")):
        logger_config.clear_loggers = True
    if (console_log_level := os.getenv("COMPRESSED_TENSORS_LOG_LEVEL")) is not None:
        logger_config.console_log_level = console_log_level.upper()
    if (log_file := os.getenv("COMPRESSED_TENSORS_LOG_FILE")) is not None:
        logger_config.log_file = log_file
    if (log_file_level := os.getenv("COMPRESSED_TENSORS_LOG_FILE_LEVEL")) is not None:
        logger_config.log_file_level = log_file_level.upper()

    if logger_config.disabled:
        logger.disable("compressed_tensors")
        return

    logger.enable("compressed_tensors")

    if logger_config.clear_loggers:
        logger.remove()

    if logger_config.console_log_level:
        # log as a human readable string with the time, function, level, and message
        logger.add(
            sys.stdout,
            level=logger_config.console_log_level.upper(),
            format="{time} | {function} | {level} - {message}",
            filter=support_log_once,
        )

    if logger_config.log_file or logger_config.log_file_level:
        log_file = logger_config.log_file or "compressed_tensors.log"
        log_file_level = logger_config.log_file_level or "INFO"
        # log as json to the file for easier parsing
        logger.add(
            log_file,
            level=log_file_level.upper(),
            serialize=True,
            filter=support_log_once,
        )


def support_log_once(record: Dict[str, Any]) -> bool:
    """
    Support logging only once using `.bind(log_once=True)`

    ```
    logger.bind(log_once=False).info("This will log multiple times")
    logger.bind(log_once=False).info("This will log multiple times")
    logger.bind(log_once=True).info("This will only log once")
    logger.bind(log_once=True).info("This will only log once")  # skipped
    ```
    """
    log_once = record["extra"].get("log_once", False)
    level = getattr(record["level"], "name", "none")
    message = hash(str(level) + record["message"])

    if log_once and message in _logged_once:
        return False

    if log_once:
        _logged_once.add(message)

    return True


# invoke logger setup on import with default values enabling console logging with INFO
# and disabling file logging
configure_logger(config=LoggerConfig())



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

from typing import List, Optional

import torch
from compressed_tensors.config import CompressionFormat
from compressed_tensors.quantization import QuantizationScheme
from compressed_tensors.quantization.utils import is_module_quantized
from loguru import logger


__all__ = ["infer_model_format", "infer_module_format"]


# Priority order for compression format matching
# More specific formats should come before more general ones
COMPRESSION_FORMAT_PRIORITY: List[CompressionFormat] = [
    CompressionFormat.mxfp4_pack_quantized,
    CompressionFormat.mxfp8_quantized,
    CompressionFormat.nvfp4_pack_quantized,
    CompressionFormat.pack_quantized,
    CompressionFormat.int_quantized,
    CompressionFormat.float_quantized,
    CompressionFormat.naive_quantized,
    CompressionFormat.dense,
]


def infer_model_format(
    model: torch.nn.Module,
    force_compression_format: Optional[str] = None,
) -> CompressionFormat:
    """
    Infers the quantization format for a model based on its modules

    For a summary of the formats, see `docs/guides/compression_formats.md`.

    :param model: model to check for quantization
    :param quantization_format: optional global format to override
        the per module formats
    :return: list of formats applied to modules (excluding dense format)
    """
    formats = set()

    for _, module in model.named_modules(remove_duplicate=True):
        if not is_module_quantized(module):
            continue

        # infer format using priority list
        scheme: QuantizationScheme = module.quantization_scheme
        format = infer_module_format(type(module), scheme)

        # user provides a global override format
        if force_compression_format is not None:
            if force_compression_format != format.value:
                logger.warning(
                    f"The provided format {force_compression_format} does not match "
                    f"the inferred format {format.value}. Compression may fail",
                    log_once=True,
                )
            format = force_compression_format

        # user provides a format via QuantizationScheme.format
        elif scheme.format is not None:
            format = scheme.format

        scheme.format = CompressionFormat(format)
        if format != CompressionFormat.dense:
            formats.add(format)

    return _flatten_formats(formats)


def infer_module_format(
    module_type: type, scheme: QuantizationScheme
) -> CompressionFormat:
    """
    Infer the module's compression format using the module's type and quant scheme

    :param module_type: module type, typically linear
    :param scheme: quantization applied to module
    :return: format that should be used to compress the module
    """
    # avoid circular imports
    from compressed_tensors.compressors import BaseCompressor

    return next(
        (
            format
            for format in COMPRESSION_FORMAT_PRIORITY
            if BaseCompressor.get_value_from_registry(format.value).can_compress(
                module_type, scheme
            )
        )
    )


def _flatten_formats(formats: set[CompressionFormat]) -> CompressionFormat:
    """
    Reduce a list of compression formats to a single summary format.

    Returns dense if the list is empty, the single format if there's only one,
    or mixed_precision if there are multiple different formats.

    :param formats: list of compression formats found in the model
    :return: single compression format representing the overall model
    """
    if len(formats) <= 0:
        return CompressionFormat.dense
    if len(formats) == 1:
        return list(formats)[0]
    if len(formats) >= 2:
        return CompressionFormat.mixed_precision



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import os
from typing import TYPE_CHECKING, Literal, Optional

import torch
from compressed_tensors.offload.cache import OffloadCache
from compressed_tensors.offload.dist_utils import is_rank0
from compressed_tensors.offload.utils import send_tensors, to_tensor
from safetensors import safe_open
from safetensors.torch import save_file


if TYPE_CHECKING:
    from torch._prims_common import DeviceLikeType


class DiskCache(OffloadCache):
    """
    Handles offloading and onloading tensors from/to disk.

    Tensors usually start as a key in safetensors file, converted by (TODO NAME).
    New or updated tensors are written to new safetensors files in `offload_dir`.

    Tensors are stored in memory as meta tensors. The mapping between offloaded meta
    tensors and their locations on disk is defined by `index`.
    """

    offload_device = "disk"

    # offloaded tensors -> weight info
    index: dict[torch.Tensor, dict[str, str]] = dict()

    # directory where new tensors are written to
    offload_dir: str
    _new_file_prefix = "ct_disk_cache"

    def __init__(
        self,
        onload_device: torch.device,
        offload_device: Optional["DeviceLikeType | Literal['disk']"] = None,
        offload_dir: Optional[str] = None,
    ):
        super().__init__(onload_device, offload_device=offload_device)
        if offload_device is not None:
            assert str(offload_device) == str(self.offload_device)

        if offload_dir is None:
            raise ValueError(
                "Must provide an `offload_dir` to perform disk offloading "
                "(add `offload_folder` argument to `from_pretrained`)"
            )
        self.offload_dir = offload_dir

    def onload(self, offloaded: torch.Tensor | None) -> torch.Tensor | None:
        """
        Onload a tensor from disk/meta to device

        :param offloaded: meta tensor to onload
        :return: device tensor, read from disk
        """
        if offloaded is None:
            return None

        weight_info = self.index[offloaded]
        device = _get_safe_open_device(self.onload_device)

        with safe_open(
            weight_info["safetensors_file"], framework="pt", device=device
        ) as file:
            onloaded = file.get_tensor(weight_info["weight_name"])
            onloaded = to_tensor(onloaded, offloaded)
            onloaded = onloaded.to(getattr(torch, weight_info["dtype"]))
            return onloaded

    def offload(
        self, tensor: torch.Tensor | None, offloaded: Optional[torch.Tensor] = None
    ) -> torch.Tensor | None:
        """
        Offload a tensor to disk by writing a new safetensors file

        :param tensor: tensor on any device
        :param offloaded: optional meta tensor used to look up an existing file
        :return: meta tensor representing the offloaded tensor
        """
        if tensor is None:
            return None

        if tensor.device.type == "meta":
            assert tensor in self.index
            return tensor

        if offloaded is None:
            offloaded = send_tensors(tensor, device="meta")

        file_name = f"{self._new_file_prefix}{id(offloaded)}.safetensors"
        file_path = os.path.join(self.offload_dir, file_name)
        self.index[offloaded] = {
            "safetensors_file": file_path,
            "weight_name": "weight",
            "dtype": str(tensor.dtype).removeprefix("torch."),
        }

        save_file({"weight": tensor}, file_path)
        return offloaded

    def __delitem__(self, key: str):
        """
        Remove the offload associated with `key`. If a new file was created to store
        updated tensor data, that new tensor data file is deleted.

        Any references to onloaded tensors held by this class are invalidated.

        :param key: name of tensor to invalidate
        """
        offloaded = self.offloaded_values[key]
        file_path = self.index[offloaded]["safetensors_file"]
        if os.path.basename(file_path).startswith(self._new_file_prefix):
            os.remove(file_path)
        del self.index[offloaded]
        super().__delitem__(key)

    def update_offload(self, offloaded: torch.Tensor, data: torch.Tensor | None):
        """
        Write new param data to file that already exists.

        :param offloaded: meta tensors representating parameter to update
        :param data: new data
        """
        # get weight info from index
        assert offloaded in self.index, "Cannot find offload to update"
        weight_info = self.index[offloaded]
        file_path = weight_info["safetensors_file"]
        weight_name = weight_info["weight_name"]
        dtype = getattr(torch, weight_info["dtype"])

        # create new file if old file was a symlink to a checkpoint file
        if os.path.islink(file_path):
            assert os.path.basename(file_path).startswith(self._new_file_prefix)
            os.unlink(file_path)

        # save with data using original weight_name
        save_file({weight_name: data.reshape_as(offloaded).to(dtype=dtype)}, file_path)

    @classmethod
    def create_checkpoint_symlink(
        cls,
        offloaded: torch.Tensor,
        weight_info: dict,
        offload_dir: str | os.PathLike | None,
    ) -> None:
        assert is_rank0(), "Must call on rank 0 to avoid id collisions between ranks"
        if offload_dir is None:
            raise ValueError(
                "Must provide an `offload_dir` to perform disk offloading "
                "(add `offload_folder` argument to `from_pretrained`)"
            )
        file_name = f"{cls._new_file_prefix}{id(offloaded)}.safetensors"
        file_path = os.path.join(offload_dir, file_name)

        os.symlink(weight_info["safetensors_file"], file_path)
        cls.index[offloaded] = {
            "safetensors_file": file_path,
            "weight_name": weight_info["weight_name"],
            "dtype": weight_info["dtype"],
        }


def _get_safe_open_device(device: "DeviceLikeType") -> str | int:
    """
    `safetensors.safe_open` does not accept `torch.device` as argument, so
    we must convert from torch.device to a string, while considering "cuda" resolution

    :param device: torch device to convert
    :return: device argument to `safetensors.safe_open`
    """
    device = torch.device(device)
    if device.type in ("cuda"):
        if device.index is None:
            return torch.cuda.current_device()
        else:
            return device.index
    else:
        return device.type



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

from typing import Protocol

import torch
from compressed_tensors.quantization import QuantizationConfig


class Converter(Protocol):
    """
    Converter interface, to modify safetensors files based on tensor name and
    pointer to torch.Tensor, and create the QuantizationConfig
    """

    def process(self, tensors: dict[str, torch.Tensor]):
        """
        Operate on safetensors file in-place, to convert it into a compressed-tensors
        compatible format.
        e.g. rename tensor, or invert weights to match compressed-tensors convention.

        :param tensors: dictionary of tensor name to tensor, as loaded from
        safetensors file. Tensor name is a concatenation of module name and
        parameter name, e.g.
        - `model.layers.0.self_attn.q_proj.weight`
        - `model.layers.0.mlp.up_proj.weight_packed`
        """
        pass

    def validate(self, tensors: dict[str, torch.Tensor]):
        """
        Validation layer to quickly log warnings or raise an error if the safetensors
        file is not compatible with Converter.

        :param tensors: dictionary of tensor name to tensor, as loaded from
        safetensors file.
        """
        pass

    def create_config(self) -> QuantizationConfig:
        """
        Create compressed-tensors QuantizationConfig so that it can be set in the
        new model checkpoint's config.json.
        """
        pass



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

from compressed_tensors.compressors.base import BaseCompressor
from compressed_tensors.config import CompressionFormat
from compressed_tensors.quantization import QuantizationScheme
from compressed_tensors.utils import TensorStateDict


__all__ = ["DenseCompressor"]


@BaseCompressor.register(name=CompressionFormat.dense.value)
class DenseCompressor(BaseCompressor):
    """
    Identity compressor for dense models — both compress and decompress
    return the state dict unchanged.
    """

    @classmethod
    def compress(
        cls, state_dict: TensorStateDict, scheme: QuantizationScheme
    ) -> TensorStateDict:
        """
        Compress a per-module state dict.

        For dense models, this is a no-op.

        :param state_dict: local-name state dict (weight, ...)
        :param scheme: quantization scheme (unused)
        :return: unmodified state dict
        """
        return state_dict

    @classmethod
    def decompress(
        cls, state_dict: TensorStateDict, scheme: QuantizationScheme
    ) -> TensorStateDict:
        """
        Decompress a per-module state dict.

        For dense models, this is a no-op.

        :param state_dict: local-name state dict (weight, ...)
        :param scheme: quantization scheme (unused)
        :return: unmodified state dict
        """
        return state_dict

    @classmethod
    def can_compress(cls, module_type: type, scheme: QuantizationScheme) -> bool:
        """Dense compressor matches when there's no weight quantization."""
        return True



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

import pytest
from compressed_tensors.quantization import QuantizationArgs, QuantizationScheme
from pydantic import ValidationError


def test_basic_scheme():
    targets = ["model.layer.0", "model.layer.3"]
    weights = QuantizationArgs()

    scheme = QuantizationScheme(targets=targets, weights=weights)
    assert scheme.targets == targets
    assert scheme.weights == weights
    assert scheme.input_activations is None
    assert scheme.output_activations is None
    assert scheme.format is None


def test_full_scheme():
    targets = ["Linear"]
    weights = QuantizationArgs()
    input_activations = QuantizationArgs(num_bits=8)
    output_activations = QuantizationArgs(num_bits=8, type="float", symmetric=False)

    scheme = QuantizationScheme(
        targets=targets,
        weights=weights,
        input_activations=input_activations,
        output_activations=output_activations,
        format="float-quantized",
    )
    assert scheme.targets == targets
    assert scheme.weights == weights
    assert scheme.input_activations == input_activations
    assert scheme.output_activations == output_activations
    assert scheme.format == "float-quantized"


def test_needs_targets():
    with pytest.raises(ValidationError):
        _ = QuantizationScheme()


def test_defaults():
    targets = ["Linear"]
    output = QuantizationScheme(targets=targets)
    assert output.weights is None
    assert output.input_activations is None
    assert output.output_activations is None
    assert output.format is None


def test_serialize_scheme():
    """Test serialization of QuantizationScheme including format"""
    from compressed_tensors.config import CompressionFormat

    targets = ["Linear"]
    weights = QuantizationArgs(num_bits=4, symmetric=True, group_size=128)
    input_activations = QuantizationArgs(num_bits=8, dynamic=True)
    output_activations = QuantizationArgs(num_bits=8, type="float", symmetric=False)

    scheme = QuantizationScheme(
        targets=targets,
        weights=weights,
        input_activations=input_activations,
        output_activations=output_activations,
        format=CompressionFormat.pack_quantized,
    )

    # Serialize to dict
    scheme_dict = scheme.model_dump()
    assert scheme_dict["targets"] == targets
    assert scheme_dict["format"] == "pack-quantized"
    assert "weights" in scheme_dict
    assert scheme_dict["weights"]["num_bits"] == 4
    assert "input_activations" in scheme_dict
    assert "output_activations" in scheme_dict

    # Deserialize from dict
    reloaded = QuantizationScheme.model_validate(scheme_dict)
    assert reloaded == scheme



# SPDX-License-Identifier: Apache-2.0
# SPDX-FileCopyrightText: Copyright contributors to the vLLM project

from functools import wraps
from math import ceil

import torch
from compressed_tensors.quantization.quant_args import (
    DynamicType,
    QuantizationArgs,
    QuantizationStrategy,
    round_to_quantized_type_args,
)
from compressed_tensors.quantization.quant_config import QuantizationStatus
from compressed_tensors.quantization.quant_scheme import QuantizationScheme
from compressed_tensors.quantization.utils import (
    calculate_range,
    compute_dynamic_scales_and_zp,
    maybe_pad_tensor_for_block_quant,
)
from compressed_tensors.utils import patch_attr
from torch.nn import Module


__all__ = [
    "quantize",
    "dequantize",
    "fake_quantize",
    "set_forward_quantized",
    "forward_quantize",
]


@torch.no_grad()
def quantize(
    x: torch.Tensor,
    scale: torch.Tensor,
    zero_point: torch.Tensor,
    args: QuantizationArgs,
    dtype: torch.dtype | None = None,
    g_idx: torch.Tensor | None = None,
    global_scale: torch.Tensor | None = None,
) -> torch.Tensor:
    """
    Quantize the input tensor x using the QuantizationStrategy specified in args.
    Quantization can be done per tensor, channel, token or group. For group
    quantization, the group_size must be divisible by the column size. The input scale
    and zero_points are reshaped to support vectorization (Assumes 1 is the
    channel dimension)

    :param x: Input tensor
    :param scale: scale tensor
    :param zero_point: zero point tensor
    :param args: quantization args dictating how to quantize x
    :param dtype: optional dtype to cast the quantized output to
    :param g_idx: optional mapping from column index to group index
    :param global_scale: optional constant to scale the quantization scale during QDQ
    :return: fake quantized tensor
    """

    return _process_quantization(
        x=x,
        scale=scale,
        zero_point=zero_point,
        args=args,
        dtype=dtype,
        do_quantize=True,
        do_dequantize=False,
        g_idx=g_idx,
        global_scale=global_scale,
    )


@torch.no_grad()
def dequantize(
    x_q: torch.Tensor,
    scale: torch.Tensor,
    zero_point: torch.Tensor | None = None,
    args: QuantizationArgs | None = None,
    dtype: torch.dtype | None = None,
    g_idx: torch.Tensor | None = None,
    global_scale: torch.Tensor | None = None,
) -> torch.Tensor:
    """
    Dequantize a quantized input tensor x_q based on the strategy specified in args. If
    args is not provided, the strategy will be inferred.

    :param x: quantized input tensor
    :param scale: scale tensor
    :param zero_point: zero point tensor
    :param args: quantization args used to quantize x_q
    :param dtype: optional dtype to cast the dequantized output to
    :param g_idx: optional mapping from column index to group index
    :param global_scale: optional constant to scale the quantization scale during QDQ
    :return: dequantized float tensor
    """
    if args is None:
        if scale.ndim == 0 or scale.ndim == 1:
            args = QuantizationArgs(strategy=QuantizationStrategy.TENSOR)
        elif scale.ndim == 2:
            if scale.shape[1] == 1:
                args = QuantizationArgs(strategy=QuantizationStrategy.CHANNEL)
            # Scale height matches input or is 1 -> group quantization across columns
            #
            # Example 1: scale.shape[0] == 1
            # x_q: (4, 8), scale: (1, 4) -> 2 columns per group
            #
            # Example 2: scale.shape[0] == x_q.shape[0]
            # x_q: (4, 8), scale: (4, 4) -> 2 elements per group (per row)
            elif (scale.shape[0] == 1) or (scale.shape[0] == x_q.shape[0]):
                group_size = int(x_q.shape[1] / scale.shape[1])
                args = QuantizationArgs(
                    strategy=QuantizationStrategy.GROUP, group_size=group_size
                )
            else:
                rows, cols = x_q.shape[-2], x_q.shape[-1]
                block_height = rows // scale.shape[0]  # Rows per block
                block_width = cols // scale.shape[1]  # Columns per block

                args = QuantizationArgs(
                    strategy=QuantizationStrategy.BLOCK,
                    block_structure=[block_height, block_width],
                )
        else:
            raise ValueError(
                f"Could not infer a quantization strategy from scale with {scale.ndim} "
                "dimmensions. Expected 0 or 2 dimmensions."
            )

    if dtype is None:
        dtype = scale.dtype

    return _process_quantization(
        x=x_q,
        scale=scale,
        zero_point=zero_point,
        args=args,
        do_quantize=False,
        do_dequantize=True,
        dtype=dtype,
        g_idx=g_idx,
        global_scale=global_scale,
    )


@torch.no_grad()
def fake_quantize(
    x: torch.Tensor,
    scale: torch.Tensor,
    zero_point: torch.Tensor,
    args: QuantizationArgs,
    g_idx: torch.Tensor | None = None,
    global_scale: torch.Tensor | None = None,
) -> torch.Tensor:
    """
    Fake quantize the input tensor x by quantizing then dequantizing with
    the QuantizationStrategy specified in args. Quantization can be done per tensor,
    channel, token or group. For group quantization, the group_size must be divisible
    by the column size. The input scale  and zero_points are reshaped to support
    vectorization (Assumes 1 is the channel dimension)

    :param x: Input tensor
    :param scale: scale tensor
    :param zero_point: zero point tensor
    :param args: quantization args dictating how to quantize x
    :param g_idx: optional mapping from column index to group index
    :param global_scale: optional constant to scale the quantization scale during QDQ
    :return: fake quantized tensor
    """
    return _process_quantization(
        x=x,
        scale=scale,
        zero_point=zero_point,
        args=args,
        do_quantize=True,
        do_dequantize=True,
        g_idx=g_idx,
        global_scale=global_scale,
    )


@torch.no_grad()
def _process_quantization(
    x: torch.Tensor,
    scale: torch.Tensor,
    zero_point: torch.Tensor,
    args: QuantizationArgs,
    g_idx: torch.Tensor | None = None,
    dtype: torch.dtype | None = None,
    do_quantize: bool = True,
    do_dequantize: bool = True,
    global_scale: torch.Tensor | None = None,
) -> torch.Tensor:
    q_min, q_max = calculate_range(args, x.device)
    group_size = args.group_size

    # blockwise FP8: quantize per 2D block, supports block_structure for static block
    # quantization
    if args.strategy == QuantizationStrategy.BLOCK:
        original_shape = x.shape
        block_height, block_width = args.block_structure

        x = maybe_pad_tensor_for_block_quant(x, args.block_structure)
        padded_shape = x.shape

        # reshape into blocks and transpose to make each block contiguous
        num_rows_blocks = padded_shape[0] // block_height
        num_cols_blocks = padded_shape[1] // block_width
        x_blocks = x.reshape(
            num_rows_blocks,
            block_height,
            num_cols_blocks,
            block_width,
        ).transpose(1, 2)

        # expand scale/zero_point for blocks
        sb = scale.unsqueeze(-1).unsqueeze(-1)
        zb = zero_point.unsqueeze(-1).unsqueeze(-1) if zero_point is not None else None
        if do_quantize and do_dequantize:
            x_blocks = _quantize_dequantize(
                x=x_blocks,
                scale=sb,
                zero_point=zb,
                q_min=q_min,
                q_max=q_max,
                args=args,
                global_scale=global_scale,
            )
        elif do_quantize:
            x_blocks = _quantize(
                x=x_blocks,
                scale=sb,
                zero_point=zb,
                q_min=q_min,
                q_max=q_max,
                args=args,
                dtype=dtype,
                global_scale=global_scale,
            )
        elif do_dequantize:
            x_blocks = _dequantize(
                x_q=x_blocks,
                scale=sb,
                zero_point=zb,
                global_scale=global_scale,
            )
        # restore padded shape
        output = x_blocks.transpose(1, 2).reshape(padded_shape)
        # truncate to original dimensions if padding was applied
        if original_shape != padded_shape:
            output = output[tuple([slice(v) for v in original_shape])]
    elif args.strategy in (
        QuantizationStrategy.GROUP,
        QuantizationStrategy.TENSOR_GROUP,
    ):
        output_dtype = dtype if dtype is not None else x.dtype
        output = torch.zeros_like(x).to(output_dtype)
        columns = output.shape[-1]

        # TODO: make validation step for inputs

        while scale.ndim < 2:
            # pad scale and zero point dims for slicing
            scale = scale.unsqueeze(1)
            zero_point = zero_point.unsqueeze(1) if zero_point is not None else None

        if columns >= group_size:
            if columns % group_size != 0:
                raise ValueError(
                    "tensor column shape must be divisble "
                    f"by the given group_size {group_size} but got {columns}"
                )

        # support column-order (default) quantization as well as other orderings
        # such as activation ordering. Below checks if g_idx has been initialized
        is_column_order = g_idx is None or g_idx.device.type == "meta" or -1 in g_idx
        if is_column_order:
            num_groups = int(ceil(columns / group_size))
            group_sizes = torch.full((num_groups,), group_size, dtype=torch.int)

        else:
            group_indices, group_sizes = torch.unique(g_idx, return_counts=True)
            group_sizes = group_sizes[torch.argsort(group_indices)]

            perm = torch.argsort(g_idx)
            x = x.index_select(-1, perm)

        # Maintain all dimensions except the last dim, which is divided by group_size
        reshaped_dims = (
            ceil(x.shape[-1] / group_size),
            group_size,
        )
        x = x.unflatten(-1, reshaped_dims)

        if do_quantize and do_dequantize:
            output = _quantize_dequantize(
                x=x,
                scale=scale.unsqueeze(-1),
                zero_point=zero_point.unsqueeze(-1) if zero_point is not None else None,
                q_min=q_min,
                q_max=q_max,
                args=args,
                global_scale=global_scale,
            )
        elif do_quantize:
            output = _quantize(
                x=x,
                scale=scale.unsqueeze(-1),
                zero_point=zero_point.unsqueeze(-1) if zero_point is not None else None,
                dtype=dtype,
                global_scale=global_scale,
                q_min=q_min,
                q_max=q_max,
                args=args,
            )
        elif do_dequantize:
            output = _dequantize(
                x_q=x,
                scale=scale.unsqueeze(-1),
                zero_point=zero_point.unsqueeze(-1) if zero_point is not None else None,
                global_scale=global_scale,
            )

        output = output.flatten(start_dim=-2)
        output = output.to(output_dtype)

        if not is_column_order:
            inv_perm = torch.argsort(perm)
            output = output.index_select(-1, inv_perm)

    else:  # covers tensor, channel, token, and attn_head strategies
        if do_quantize and do_dequantize:
            output = _quantize_dequantize(
                x=x,
                scale=scale,
                zero_point=zero_point,
                q_min=q_min,
                q_max=q_max,
                args=args,
                global_scale=global_scale,
            )
        elif do_quantize:
            output = _quantize(
                x=x,
                scale=scale,
                zero_point=zero_point,
                q_min=q_min,
                q_max=q_max,
                args=args,
                dtype=dtype,
                global_scale=global_scale,
            )
        elif do_dequantize:
            output = _dequantize(
                x,
                scale=scale,
                zero_point=zero_point,
                global_scale=global_scale,
            )

    return output


def set_forward_quantized(module: torch.nn.Linear | torch.nn.Embedding):
    """
    Replace a linear or embedding module's forward function with one that performs
    on-the-fly QDQ. Note that weight quantiation will be skipped for compressed modules.

    All QDQ operations can be skipped by setting `module.quantization_enabled = False`

    :param module: linear or embedding module whose forward function will be replaced
    """

    @wraps(module.forward.__func__)
    def quantized_forward(
        self: torch.nn.Linear | torch.nn.Embedding, input: torch.Tensor
    ) -> torch.Tensor:
        """
        Quantized forward pass of a linear or embedding module

        :param self: instance of linear or embedding module
        :param input: input activations to this module
        :return: linear or embedding output
        """
        scheme: QuantizationScheme | None = getattr(self, "quantization_scheme", None)
        status: QuantizationStatus | None = getattr(self, "quantization_status", None)
        enabled: bool = (
            getattr(self, "quantization_enabled", True)
            and scheme is not None
            and status is not None
        )
        weight = self.weight  # onload once
        weight_data = weight.data

        if enabled and scheme.input_activations:
            input = forward_quantize(self, input, "input", scheme.input_activations)

        if enabled and scheme.weights and status < QuantizationStatus.COMPRESSED:
            weight_data = forward_quantize(self, weight_data, "weight", scheme.weights)

        with patch_attr(weight, "data", weight_data):
            output = self.__class__.forward(self, input)

        if enabled and scheme.output_activations:
            output = forward_quantize(self, output, "output", scheme.output_activations)

        return output

    module.forward = quantized_forward.__get__(module)


def forward_quantize(
    module: Module, value: torch.Tensor, base_name: str, args: "QuantizationArgs"
) -> torch.Tensor:
    # in compressed mode, the weight is already compressed and quantized so we don't
    # need to run fake quantization
    # TODO: remove this line, as this is already guarded by `set_forward_quantized`
    if (
        module.quantization_status >= QuantizationStatus.COMPRESSED
        and base_name == "weight"
    ):
        return value

    if value.numel() == 0:
        # if the tensor is empty,
        # skip quantization
        return value

    g_idx = getattr(module, "weight_g_idx", None)
    global_scale = getattr(module, f"{base_name}_global_scale", None)

    if args.dynamic in (True, DynamicType.LOCAL):
        # dynamic quantization - determine the scale/zp on the fly
        scale, zero_point = compute_dynamic_scales_and_zp(
            value=value, args=args, module=module, global_scale=global_scale
        )
    else:
        # static quantization - get scale and zero point from layer
        scale = getattr(module, f"{base_name}_scale")
        zero_point = getattr(module, f"{base_name}_zero_point", None)

    return fake_quantize(
        x=value,
        scale=scale,
        zero_point=zero_point,
        args=args,
        g_idx=g_idx,
        global_scale=global_scale,
    )


@torch.no_grad()
def _quantize_dequantize(
    x: torch.Tensor,
    scale: torch.Tensor,
    zero_point: torch.Tensor | None,
    q_min: torch.Tensor,
    q_max: torch.Tensor,
    args: QuantizationArgs,
    global_scale: torch.Tensor | None = None,
) -> torch.Tensor:
    """
    Fused quantize-then-dequantize in a single pass, avoiding:
    - Double scale/global_scale division
    - Intermediate quantized dtype allocation
    """
    # compute effective scale once
    if global_scale is not None:
        scale = scale / global_scale

    scaled = x / scale

    if zero_point is not None:
        scaled += zero_point.to(x.dtype)

    # clamp and round (stays in float — no int8/fp8 intermediate)
    quantized = round_to_quantized_type_args(
        tensor=scaled, args=args, min=q_min, max=q_max
    )

    # dequantize: subtract zero_point and multiply by scale
    # cast to scale.dtype to match _dequantize behavior
    dequant = quantized.to(scale.dtype)
    if zero_point is not None:
        dequant = dequant - zero_point.to(scale.dtype)

    return dequant * scale


@torch.no_grad()
def _quantize(
    x: torch.Tensor,
    scale: torch.Tensor,
    zero_point: torch.Tensor,
    q_min: torch.Tensor,
    q_max: torch.Tensor,
    args: QuantizationArgs,
    dtype: torch.dtype | None = None,
    global_scale: torch.Tensor | None = None,
) -> torch.Tensor:
    # if a global scale is optionally provided, use it
    # to further scale the local `scale` parameter
    if global_scale is not None:
        scale = scale / global_scale

    scaled = x / scale

    if zero_point is not None:
        scaled += zero_point.to(x.dtype)

    # clamp and round
    quantized_value = round_to_quantized_type_args(
        tensor=scaled, args=args, min=q_min, max=q_max
    )

    if dtype is not None:
        quantized_value = quantized_value.to(dtype)

    return quantized_value


@torch.no_grad()
def _dequantize(
    x_q: torch.Tensor,
    scale: torch.Tensor,
    zero_point: torch.Tensor | None = None,
    dtype: torch.dtype | None = None,
    global_scale: torch.Tensor | None = None,
) -> torch.Tensor:
    # if a global scale is optionally provided, use it
    # to further scale the local `scale` parameter
    if global_scale is not None:
        scale = scale / global_scale

    dequant_value = x_q.to(scale.dtype)

    if zero_point is not None:
        dequant_value = dequant_value - zero_point.to(scale.dtype)

    dequant_value = dequant_value * scale

    if dtype is not None:
        dequant_value = dequant_value.to(dtype)

    return dequant_value
